(function() {
  // backtalk.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // This is the client-side backchannel support, which allows a
  // client to handshake to the server, request a reverse-proxy listen
  // port, and perform other sideways interactions with the server.
  var _, async, checkPass0, config, connect, createClientVitae, crypto, events, forge, fs, handshake, mkdirp, net, os, patches, path, pingPass0, registerPass0, registerPass1, registerPass2, registerPass3, registerPass4, streams, toml, unregisterPass0, unregisterPass1, unregisterPass2, upgradePass0, util;

  _ = require('lodash');

  async = require('async');

  crypto = require('crypto');

  events = require('events');

  forge = require('node-forge');

  fs = require('fs');

  mkdirp = require('mkdirp');

  net = require('net');

  os = require('os');

  path = require('path');

  patches = require('./patches');

  util = require('./util');

  streams = require('../streams');

  config = require('../config');

  toml = require('toml');

  createClientVitae = function(cfgData) {
    var ref, vitae;
    vitae = _.cloneDeep((ref = cfgData.vitae) != null ? ref : {});
    _.assign(vitae, {
      hostname: os.hostname(),
      arch: os.arch(),
      platform: os.platform(),
      release: os.release(),
      version: config.version().version
    });
    return vitae;
  };

  exports.create = function() {
    var protocol;
    protocol = new events.EventEmitter();
    protocol.connect = function(lt) {
      return connect(protocol, lt, true);
    };
    protocol.open = function(cfgData) {
      return connect(protocol, {
        config: cfgData
      }, false);
    };
    return protocol;
  };

  connect = function(protocol, lt, doShake) {
    var conn, ended, handler;
    protocol.status = {};
    protocol.config = lt.config;
    _.defaults(protocol.config, {
      host: 'localhost',
      backPort: 3014
    });
    ended = false;
    protocol.end = function() {};
    protocol.connection = handler = (function() {
      var handle, handlerStack, io, push, readAccumulate, write;
      handlerStack = [];
      readAccumulate = '';
      return {
        push: push = function(handler) {
          if (!_.isFunction(handler)) {
            return protocol.emit('error', `Non-functional handler ${handler}`);
          } else {
            return handlerStack.push(handler);
          }
        },
        write: write = function(data) {
          return conn.write(JSON.stringify(data) + '\x03');
        },
        draw: function(sdata) {
          var etx, rdata, results;
          readAccumulate += sdata;
          results = [];
          while ((etx = readAccumulate.indexOf('\x03')) !== -1) {
            rdata = readAccumulate.substr(0, etx);
            readAccumulate = readAccumulate.substr(etx + 1);
            results.push(handle(rdata));
          }
          return results;
        },
        handle: handle = function(sdata) {
          var data, e;
          try {
            data = JSON.parse(sdata);
          } catch (error) {
            e = error;
            return protocol.emit('error', `JSON parse error: ${e}`);
          }
          if (handlerStack.length === 0) {
            return protocol.emit('error', "No handler left");
          }
          return handlerStack.shift()(data);
        },
        io: io = function(out, handler) {
          push(handler);
          return write(out);
        },
        expect: function(out, expectedType, handler, errHandler) {
          return io(out, function(data) {
            if (data.type === 'error') {
              protocol.emit('error', data.error);
              return typeof errHandler === "function" ? errHandler(data) : void 0;
            } else if (data.type !== expectedType) {
              protocol.emit('error', `Expected ${expectedType}, got ${data.type}`);
              return typeof errHandler === "function" ? errHandler(data) : void 0;
            } else {
              return handler(data);
            }
          });
        },
        handleRaw: function(cb) {
          conn.removeListener('data', handler.draw);
          conn.pause();
          // see https://github.com/nodejs/node/issues/6038
          conn.setEncoding(null);
          delete conn._readableState.decoder;
          return cb(conn, function() {
            conn.setEncoding('utf8');
            conn.on('data', handler.draw);
            return conn.resume();
          });
        }
      };
    })();
    conn = new net.Socket();
    conn.setEncoding('utf8');
    conn.on('error', function(err) {
      return protocol.emit('error', `Conn err: ${err} ${err.stack}`);
    });
    conn.on('data', handler.draw);
    conn.on('close', function(hadErr) {
      return protocol.emit('close', hadErr);
    });
    return conn.connect(protocol.config.backPort, protocol.config.host, function() {
      protocol.end = function() {
        if (!ended) {
          conn.end();
        }
        return ended = true;
      };
      protocol.register = function(tname) {
        return registerPass0(protocol, handler, tname);
      };
      protocol.unregister = function(regId) {
        return unregisterPass0(protocol, handler, regId);
      };
      protocol.ping = function(cb) {
        return pingPass0(protocol, handler, cb);
      };
      protocol.check = function(cb) {
        return checkPass0(protocol, handler, cb);
      };
      protocol.upgrade = function(cb, dstat) {
        return upgradePass0(protocol, handler, cb, dstat);
      };
      protocol.emit('connected');
      if (doShake) {
        return handshake(protocol, handler);
      }
    });
  };

  handshake = function(protocol, handler) {
    var cfgData, key, receive0, receive1, receive2, regIds;
    cfgData = protocol.config;
    regIds = _.map(cfgData.registration, function(r) {
      return r.id;
    });
    key = cfgData.keydata;
    receive0 = function() {
      var send1;
      send1 = {
        type: 'handshake1',
        handshakeVersion: 1,
        registeredId: regIds
      };
      return handler.expect(send1, 'shakehand1', receive1);
    };
    receive1 = function(data) {
      var err, resp, send2;
      try {
        resp = key != null ? crypto.privateDecrypt(key, Buffer.from(data.prompt, 'base64')) : Buffer.from(data.prompt, 'base64');
      } catch (error) {
        err = error;
        protocol.emit('error', "Unable to decrypt server prompt. (Is there a key mismatch?)");
        return protocol.end();
      }
      send2 = {
        type: 'handshake2',
        handshakeVersion: 1,
        registeredId: regIds,
        response: resp.toString('base64')
      };
      return handler.expect(send2, 'shakehand2', receive2);
    };
    receive2 = function(data) {
      var bindPort, ep, i, len, ref, ref1, reg;
      if (cfgData.server == null) {
        cfgData.server = {};
      }
      cfgData.server.vitae = data.vitae;
      bindPort = data.bindPort;
      if (!(_.isArray(bindPort))) {
        bindPort = [bindPort];
      }
      if ((ep = (ref = cfgData.server.vitae) != null ? ref.endpoint : void 0) != null) {
        cfgData.server.vitae.endpoints = {};
        ref1 = cfgData.registration;
        for (i = 0, len = ref1.length; i < len; i++) {
          reg = ref1[i];
          (function(reg) {
            var ep1, epath, epath2;
            ep1 = ep.replace(/REGID/g, reg.id);
            cfgData.server.vitae.endpoints[reg.id] = ep1;
            reg.endpoint = ep1;
            epath = path.join(reg.dir, 'endpoint');
            fs.writeFile(epath, `${ep1}\n`, {
              encoding: 'utf8'
            }, function(err) {
              return protocol.emit('save', err, epath);
            });
            epath2 = path.join(path.dirname(cfgData.configPath), 'endpoint');
            return fs.writeFile(epath2, `${ep1}\n`, {
              encoding: 'utf8'
            }, function(err) {
              return protocol.emit('save', err, epath2);
            });
          })(reg);
        }
      }
      if (cfgData.registration.length !== bindPort.length) {
        protocol.emit('error', `length mismatch: ${bindPort.length}, ${cfgData.registration.length}`);
        return protocol.end();
      }
      return async.map(_.zip(cfgData.registration, bindPort), function([r, bp], cb) {
        var tCfgPath, tenant;
        tenant = _.assign(_.clone(r), {
          bindPort: bp
        });
        tCfgPath = path.join(r.dir, 'config.toml');
        return fs.exists(tCfgPath, function(exists) {
          if (exists) {
            return fs.readFile(tCfgPath, {
              encoding: 'utf8'
            }, function(err, tCfgData) {
              var e, obj;
              if (err != null) {
                return cb(err);
              }
              try {
                obj = toml.parse(tCfgData);
              } catch (error) {
                e = error;
                return cb(e);
              }
              tenant.config = obj;
              return cb(null, tenant);
            });
          } else {
            return cb(null, tenant);
          }
        });
      }, function(err, tenants) {
        var send3, vitae;
        if (err != null) {
          protocol.emit('error', `Bad tenant configuration: ${err}`);
          return protocol.end();
        }
        vitae = createClientVitae(cfgData);
        send3 = {
          type: 'handshake3',
          vitae: vitae
        };
        protocol.status.handshake = {
          tenants: tenants
        };
        handler.write(send3); // final ack
        return protocol.emit('handshake');
      });
    };
    return receive0();
  };

  registerPass0 = function(protocol, handler, tname) {
    var cfgData;
    cfgData = protocol.config;
    return config.unregisteredTenants(protocol.config, function(err, tenPath, opts) {
      var index;
      if (err != null) {
        return protocol.emit('error', err);
      }
      if (tname) {
        index = opts.indexOf(tname);
        if (index > -1) {
          return registerPass1(protocol, handler, path.join(tenPath, opts[index], 'registration'));
        } else {
          return protocol.emit('error', `No unregistered tenant ${tname}.`);
        }
      } else {
        if (opts.length > 0) {
          return registerPass1(protocol, handler, path.join(tenPath, opts[0], 'registration'));
        } else {
          return config.lease(cfgData, function(err, newten) {
            if (err != null) {
              return protocol.emit('error', err);
            }
            return registerPass1(protocol, handler, path.join(newten, 'registration'));
          });
        }
      }
    });
  };

  registerPass1 = function(protocol, handler, regPath) {
    var cfgData, keyPath;
    cfgData = protocol.config;
    keyPath = path.join(path.dirname(cfgData.configPath), 'key');
    return fs.exists(keyPath, function(exists) {
      if (exists) {
        return fs.readFile(keyPath, {
          encoding: 'utf8'
        }, function(err, keydata) {
          if (err != null) {
            return protocol.emit('error', err);
          }
          return registerPass2(protocol, handler, regPath, keyPath, keydata);
        });
      } else {
        return registerPass2(protocol, handler, regPath, keyPath);
      }
    });
  };

  registerPass2 = function(protocol, handler, regPath, keyPath, keyData) {
    var nmDir, primeScript, prv, pub;
    if (keyData == null) {
      protocol.emit('keygen start');
      // using forge workers is wonky in an electron renderer
      nmDir = path.dirname(path.dirname(path.dirname(__filename)));
      if (path.basename(path.dirname(nmDir)) === 'node_modules') {
        nmDir = path.dirname(nmDir);
      } else {
        nmDir = path.join(nmDir, 'node_modules');
      }
      primeScript = path.join(nmDir, 'node-forge', 'js', 'prime.worker.js');
      return forge.pki.rsa.generateKeyPair({
        bits: 2048,
        workers: -1,
        workerScript: primeScript
      }, function(err, pair) {
        var privatePem;
        if (err != null) {
          return protocol.emit('error', err);
        }
        privatePem = forge.pki.privateKeyToPem(pair.privateKey);
        return fs.writeFile(keyPath, privatePem, {
          encoding: 'utf8'
        }, function(err) {
          if (err != null) {
            return protocol.emit('error', err);
          }
          return fs.chmod(keyPath, 0o600, function(err) {
            if (err != null) {
              return protocol.emit('error', err);
            }
            protocol.emit('keygen end');
            return registerPass3(protocol, handler, regPath, pair);
          });
        });
      });
    } else {
      prv = forge.pki.privateKeyFromPem(keyData);
      pub = forge.pki.setRsaPublicKey(prv.n, prv.e);
      return registerPass3(protocol, handler, regPath, {
        publicKey: pub,
        privateKey: prv
      });
    }
  };

  registerPass3 = function(protocol, handler, regPath, pair) {
    var cfgData, prefPath;
    cfgData = protocol.config;
    prefPath = path.join(path.dirname(regPath), 'preferred');
    return fs.exists(prefPath, function(exists) {
      if (exists) {
        return fs.readFile(prefPath, {
          encoding: 'utf8'
        }, function(err, data) {
          if (err != null) {
            return protocol.emit('error', err);
          }
          return registerPass4(protocol, handler, regPath, pair, prefPath, JSON.parse(data.trim()));
        });
      } else {
        return registerPass4(protocol, handler, regPath, pair, prefPath, null);
      }
    });
  };

  registerPass4 = function(protocol, handler, regPath, pair, prefPath, pref) {
    var receive0, receive1;
    receive0 = function() {
      var send1;
      send1 = {
        type: 'register1',
        handshakeVersion: 1,
        publicKey: forge.pki.publicKeyToPem(pair.publicKey),
        preferred: pref
      };
      return handler.expect(send1, 'didRegister1', receive1);
    };
    receive1 = function(data) {
      var regData, regId, regPort;
      regId = data.registeredId;
      regPort = data.registeredPort;
      regData = JSON.stringify({
        id: regId,
        port: regPort
      });
      return fs.writeFile(regPath, regData, {
        encoding: 'utf8'
      }, function(err) {
        var prefData;
        if (err != null) {
          return protocol.emit('error', err);
        }
        prefData = JSON.stringify({
          id: regId,
          port: regPort
        });
        return fs.writeFile(prefPath, `${prefData}\n`, {
          encoding: 'utf8'
        }, function(err) {
          if (err != null) {
            return protocol.emit('error', err);
          }
          handler.write({
            type: 'register2' // final ack
          });
          return protocol.emit('registered', regPath);
        });
      });
    };
    return receive0();
  };

  unregisterPass0 = function(protocol, handler, regId) {
    var cfgData, tenPath;
    cfgData = protocol.config;
    tenPath = path.join(path.dirname(cfgData.configPath), 'tenants');
    return mkdirp(tenPath).then((err) => {
      if (err != null) {
        return protocol.emit('error', err);
      }
      return fs.readdir(tenPath, function(err, dirs) {
        var regObj;
        regObj = null;
        return async.filter(dirs, function(dir, cb) {
          var regPath;
          regPath = path.join(tenPath, dir, 'registration');
          return fs.exists(regPath, function(exists) {
            if (!exists) {
              return cb(null, false);
            }
            return fs.readFile(regPath, {
              encoding: 'utf8'
            }, function(err, data) {
              var e, obj;
              if (err != null) {
                return protocol.emit('error', err);
              }
              try {
                obj = JSON.parse(data.trim());
              } catch (error) {
                e = error;
                return protocol.emit('error', e);
              }
              if (obj.id === regId) {
                regObj = obj;
                return cb(null, true);
              } else {
                return cb(null, false);
              }
            });
          });
        }, function(err, opts) {
          if (err != null) {
            return protocol.emit('error', err);
          }
          if (opts.length > 1) {
            return protocol.emit('error', `Too many regs found for ${regId}.`);
          } else if (opts.length < 1) {
            return protocol.emit('error', `No tenant found with id ${regId}.`);
          } else {
            return unregisterPass1(protocol, handler, regObj, path.join(tenPath, opts[0], 'registration'));
          }
        });
      });
    });
  };

  unregisterPass1 = function(protocol, handler, regdata, regPath) {
    var cfgData, keyPath;
    cfgData = protocol.config;
    keyPath = path.join(path.dirname(cfgData.configPath), 'key');
    return fs.readFile(keyPath, {
      encoding: 'utf8'
    }, function(err, keydata) {
      if (err != null) {
        return protocol.emit('error', err);
      }
      return unregisterPass2(protocol, handler, regdata, regPath, keydata);
    });
  };

  unregisterPass2 = function(protocol, handler, regdata, regPath, key) {
    var cfgData, receive0, receive1, receive1err, receive2, receive2err, regId, unlinkRegKey;
    cfgData = protocol.config;
    regId = regdata.id;
    receive0 = function() {
      var send1;
      send1 = {
        type: 'unregister1',
        handshakeVersion: 1,
        registeredId: regId
      };
      return handler.expect(send1, 'didUnregister1', receive1, receive1err);
    };
    receive1 = function(data) {
      var err, resp, send2;
      try {
        resp = crypto.privateDecrypt(key, Buffer.from(data.prompt, 'base64'));
      } catch (error) {
        err = error;
        protocol.emit('error', "Unable to decrypt server prompt. (Is there a key mismatch?): removing registration anyways.");
        // this might leave the server with an orphan registration, but it
        // puts the client in a better state
        unlinkRegKey();
      }
      send2 = {
        type: 'unregister2',
        handshakeVersion: 1,
        registeredId: regId,
        response: resp.toString('base64')
      };
      return handler.expect(send2, 'didUnregister2', receive2, receive2err);
    };
    receive1err = function(data) {
      return unlinkRegKey();
    };
    unlinkRegKey = function() {
      return fs.unlink(regPath, function(err) {
        var keyPath;
        if (err != null) {
          protocol.emit('error', new Error(`Couldn't remove file \"${regPath}\".`));
          return protocol.end();
        }
        keyPath = path.join(path.dirname(cfgData.configPath), 'key');
        return config.registeredTenants(cfgData, function(err, tenPath, dirs) {
          if (dirs.length === 0) {
            return fs.unlink(keyPath, function(err) {
              if (err != null) {
                protocol.emit('error', new Error(`Couldn't remove file \"${keyPath}\".`));
                return protocol.end();
              }
              handler.write({
                type: 'unregister3' // final ack
              });
              return protocol.emit('unregistered', regPath);
            });
          } else {
            handler.write({
              type: 'unregister3' // final ack
            });
            return protocol.emit('unregistered', regPath);
          }
        });
      });
    };
    receive2 = function(data) {
      return unlinkRegKey();
    };
    receive2err = function(data) {
      return unlinkRegKey();
    };
    return receive0();
  };

  pingPass0 = function(protocol, handler, cb) {
    var receive0, receive1;
    receive0 = function() {
      var send1;
      send1 = {
        type: 'ping1'
      };
      return handler.io(send1, receive1);
    };
    receive1 = function(data) {
      return cb(null, data);
    };
    return receive0();
  };

  checkPass0 = function(protocol, handler, cb) {
    var cfgData, receive0, receive1;
    cfgData = protocol.config;
    receive0 = function() {
      var send1, vitae;
      vitae = createClientVitae(cfgData);
      send1 = {
        type: 'check1',
        vitae: vitae
      };
      return handler.expect(send1, 'didCheck1', receive1);
    };
    receive1 = function(data) {
      return cb(null, data);
    };
    return receive0();
  };

  upgradePass0 = function(protocol, handler, cb, dstat) {
    var cfgData, receive0, receive1;
    cfgData = protocol.config;
    receive0 = function() {
      var send1, vitae;
      vitae = createClientVitae(cfgData);
      send1 = {
        type: 'check1',
        vitae: vitae
      };
      return handler.expect(send1, 'didCheck1', receive1);
    };
    receive1 = function(data) {
      var patch, send2, size;
      if (!data.patch.update) {
        return cb(null, {
          size: 0
        });
      }
      patch = data.patch;
      size = data.patch.size;
      if (size == null) {
        return cb(new Error('No size found'));
      }
      if (size === 0) {
        return cb(null, {
          size: size
        });
      }
      handler.handleRaw(function(conn, done) {
        return patches.makeDest(cfgData, function(err, dest) {
          var outstream, taker, tdone;
          if (err != null) {
            return cb(err);
          }
          outstream = dest.outstream();
          tdone = function() {
            return outstream.end(function() {
              done();
              return cb(null, {
                patch: patch,
                path: dest.path()
              });
            });
          };
          taker = streams.take(conn, size, tdone, dstat);
          return taker.pipe(outstream);
        });
      });
      send2 = {
        type: 'upgrade1',
        version: data.patch.version,
        id: data.patch.id
      };
      return handler.write(send2);
    };
    return receive0();
  };

}).call(this);
