(function() {
  // client.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // The top-level client daemon execution, which opens a listener port for
  // receiving commands, as well as opens up the backchannel and main ssh
  // reverse proxy tunnel.
  var backtalk, connectBacktalk, connectListener, connectTunnel, disconnectBacktalk, doListen, emitter, events, forge, listen, resetProtocol, start, tunnel, waitfor;

  forge = require('node-forge');

  events = require('events');

  emitter = require('../emitter');

  listen = require('../listen');

  waitfor = require('../waitfor');

  backtalk = require('./backtalk');

  tunnel = require('./tunnel');

  resetProtocol = function(protocol) {
    // stub out not-yet-listening functions
    return protocol.backchannel.ping = function(cb) {
      return cb("Ping not available");
    };
  };

  exports.create = function(daemon) {
    var protocol, resetHandle, resetRetry, retryDelays, retryHandle, retryIndex, retryWith;
    protocol = new events.EventEmitter();
    emitter.augment(protocol);
    protocol.daemon = daemon;
    protocol.listener = new events.EventEmitter();
    protocol.backchannel = new events.EventEmitter();
    protocol.tunnel = new events.EventEmitter();
    resetProtocol(protocol);
    resetHandle = null;
    retryHandle = null;
    retryDelays = [2, 2, 2, 2, 2, 5, 5, 5, 5, 5, 5, 15, 15, 15, 15, 15, 15, 120, 120, 120, 120, 120, 120, 120, 120, 120, 120, 300, 300, 300, 300, 300, 300, 300];
    retryIndex = 0;
    retryWith = function(configData) {
      protocol._running = false;
      if (resetHandle != null) {
        clearTimeout(resetHandle);
        resetHandle = null;
      }
      if (!protocol._wanting) {
        if (retryHandle != null) {
          clearTimeout(retryHandle);
          return retryHandle = null;
        }
      } else {
        protocol.emit('willRetry', retryIndex, retryDelays[retryIndex]);
        return retryHandle = setTimeout((function() {
          protocol.emit('retry', retryIndex);
          retryHandle = null;
          retryIndex = Math.min(retryIndex + 1, retryDelays.length - 1);
          if (protocol._wanting) {
            return protocol.start(configData);
          }
        }), retryDelays[retryIndex] * 1000);
      }
    };
    resetRetry = function() {
      if (resetHandle != null) {
        clearTimeout(resetHandle);
      }
      if (retryHandle != null) {
        clearTimeout(retryHandle);
      }
      resetHandle = null;
      retryHandle = null;
      return retryIndex = 0;
    };
    protocol._running = false;
    protocol.start = function(configData, cb) {
      var onstart;
      if (protocol._running) {
        return typeof cb === "function" ? cb() : void 0;
      }
      protocol._running = true;
      protocol._wanting = true;
      onstart = function() {
        resetHandle = setTimeout(resetRetry, 5000);
        return typeof cb === "function" ? cb() : void 0;
      };
      protocol.once('start', onstart);
      protocol.once('close', function() {
        resetProtocol(protocol);
        protocol.removeListener('start', onstart);
        return retryWith(configData);
      });
      return start(protocol, configData);
    };
    protocol.end = function(cb) {
      protocol._wanting = false;
      resetRetry();
      if (!protocol._running) {
        return typeof cb === "function" ? cb() : void 0;
      }
      if (cb != null) {
        protocol.once('close', cb);
      }
      return protocol.emit('ending');
    };
    protocol.restart = function(configData, cb) {
      return protocol.end(function() {
        return protocol.start(configData, cb);
      });
    };
    protocol._listening = false;
    protocol.listen = function(configData, cb) {
      var onlistening;
      if (protocol._listening) {
        return typeof cb === "function" ? cb() : void 0;
      }
      protocol.config = configData;
      protocol._listening = true;
      onlistening = function() {
        return typeof cb === "function" ? cb() : void 0;
      };
      protocol.once('listening', onlistening);
      protocol.once('deaf', function() {
        protocol.removeListener('listening', onlistening);
        return protocol._listening = false;
      });
      return doListen(protocol, configData);
    };
    protocol.deafen = function(cb) {
      if (!protocol._listening) {
        return typeof cb === "function" ? cb() : void 0;
      }
      if (cb != null) {
        protocol.once('deaf', cb);
      }
      return protocol.emit('deafening');
    };
    protocol.relisten = function(configData, cb) {
      return protocol.deafen(function() {
        return protocol.listen(configData, cb);
      });
    };
    protocol.status = function() {
      var configData, fingerprint, privateKey, ref, ref1, ref2, serverVitae, status, vitae;
      if (protocol._listening) {
        if (protocol._running) {
          configData = protocol.config;
          vitae = (ref = configData.vitae) != null ? ref : {};
          serverVitae = (ref1 = (ref2 = configData.server) != null ? ref2.vitae : void 0) != null ? ref1 : {};
          if (configData.keydata != null) {
            privateKey = forge.pki.privateKeyFromPem(configData.keydata);
            fingerprint = forge.ssh.getPublicKeyFingerprint(privateKey, {
              encoding: 'hex',
              delimiter: ':'
            });
          }
          status = {
            status: 'running',
            registration: configData.registration,
            fingerprint: fingerprint,
            vitae: vitae,
            server: {
              vitae: serverVitae
            }
          };
          return status;
        } else {
          return {
            status: 'listening',
            vitae,
            server: {
              vitae: serverVitae
            }
          };
        }
      } else {
        return {
          status: 'stopped'
        };
      }
    };
    return protocol;
  };

  start = function(protocol, configData) {
    var waiter;
    waiter = waitfor.create(function(isErr) {
      return protocol.emit('close', isErr);
    });
    disconnectBacktalk(protocol, protocol.listener.lt, waiter);
    return connectBacktalk(protocol, configData, protocol.listener.lt, waiter);
  };

  doListen = function(protocol, configData) {
    var waiter;
    waiter = waitfor.create(function(isErr) {
      return protocol.emit('deaf', isErr);
    });
    return connectListener(protocol, configData, waiter);
  };

  connectListener = function(protocol, configData, waiter) {
    var lt;
    lt = listen.create();
    emitter.augment(lt);
    protocol.listener.lt = lt;
    lt.on('error', function(err) {
      protocol.listener.emit('error', err);
      protocol.emit('error', err);
      return waiter.error(err);
    });
    lt.on('close', function() {
      protocol.listener.emit('close');
      return waiter.close();
    });
    lt.on('connection', function(ltc) {
      return protocol.listener.emit('connection', ltc);
    });
    lt.on('command', function(cmd, ltc) {
      protocol.listener.emit('command', cmd, ltc);
      return protocol.emit('command', cmd, ltc);
    });
    lt.on('listening', function(addr) {
      waiter.open();
      protocol.listener.emit('listening', addr);
      return protocol.emit('listening', addr);
    });
    if (protocol._listening) {
      protocol.on('deafening', function() {
        return lt.end();
      });
      return lt.listen(configData, 3015);
    } else {
      return waiter.close();
    }
  };

  disconnectBacktalk = function(protocol, lt, waiter) {
    protocol.unlisten();
    return lt.unlisten();
  };

  connectBacktalk = function(protocol, configData, lt, waiter) {
    var bt;
    bt = backtalk.create();
    protocol.backchannel.ping = function(cb) {
      if (bt.ping != null) {
        return bt.ping(cb);
      } else {
        return cb("Ping not available");
      }
    };
    bt.on('error', function(err) {
      protocol.backchannel.emit('error', err);
      protocol.emit('error', err);
      waiter.error(err);
      return bt.end();
    });
    bt.on('close', function(hadErr) {
      protocol.backchannel.emit('close', hadErr);
      return waiter.close();
    });
    bt.on('connected', function() {
      return protocol.backchannel.emit('connected');
    });
    bt.on('save', function(err, msg) {
      return protocol.backchannel.emit('save', err, msg);
    });
    bt.on('handshake', function() {
      var i, len, ref, ref1, ref2, results, tenant;
      protocol.backchannel.emit('handshake');
      waiter.open();
      ref2 = (ref = (ref1 = bt.status.handshake) != null ? ref1.tenants : void 0) != null ? ref : [];
      results = [];
      for (i = 0, len = ref2.length; i < len; i++) {
        tenant = ref2[i];
        results.push(connectTunnel(protocol, lt, bt, tenant, waiter.branch()));
      }
      return results;
    });
    lt.listenOn('close', function() {
      return bt.end();
    });
    if (protocol._running) {
      protocol.listenOn('ending', function() {
        return bt.end();
      });
      return bt.connect(lt);
    } else {
      return waiter.close();
    }
  };

  connectTunnel = function(protocol, lt, bt, tenant, waiter) {
    var ct;
    ct = tunnel.create();
    ct.on('error', function(err) {
      protocol.tunnel.emit('error', err);
      protocol.emit('error', err);
      waiter.error(err);
      return ct.end();
    });
    ct.on('close', function() {
      protocol.tunnel.emit('close');
      bt.end();
      return waiter.close();
    });
    ct.on('received', function(ctc) {
      return protocol.tunnel.emit('received', ctc);
    });
    bt.on('close', function() {
      return ct.end();
    });
    lt.listenOn('close', function() {
      return ct.end();
    });
    ct.on('connected', function() {
      protocol.tunnel.emit('connected');
      return waiter.open();
    });
    protocol.emit('start');
    if (protocol._running) {
      protocol.listenOn('ending', function() {
        return ct.end();
      });
      return ct.connect(bt, tenant);
    } else {
      return waiter.close();
    }
  };

}).call(this);
