(function() {
  // healer.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // The client self-healing process, which resets a client connection,
  // re-registering a client with a new port number (and maybe
  // a new registration ID when necessary).

  // TODO: killing other Node.js (or maybe just 'node petit-clientd')
  // processes

  // TODO: auto re-registering tenants that were previously registered
  var _, async, config, doHeal, editPreferred, events, fs, manage, path, postDelete, postKill, postSaved, postUnload, postUnregister, psnode, regIdResetTry, util;

  _ = require('lodash');

  async = require('async');

  config = require('../config');

  events = require('events');

  fs = require('fs');

  manage = require('./manage');

  path = require('path');

  util = require('./util');

  psnode = require('ps-node');

  // After a certain number of attempts, we won't bother trying to keep the
  // registration ID
  regIdResetTry = 30000;

  // Heals a connection by requesting a new socket (and maybe a new regID)
  exports.heal = function(cfgData, daemon) {
    var output, recovPath;
    output = daemon.output;
    output.log("Trying to heal connection.");
    recovPath = path.join(path.dirname(cfgData.configPath), 'healing');
    return fs.exists(recovPath, function(exists) {
      if (!exists) {
        return doHeal(cfgData, daemon, recovPath);
      }
      return fs.readFile(recovPath, {
        encoding: 'utf8'
      }, function(err, data) {
        var e, now, recovData;
        if (err != null) {
          output.warn(`Error reading ${recovPath}: proceeding anyways`);
          return doHeal(cfgData, daemon, recovPath);
        }
        try {
          recovData = JSON.parse(data.trim());
        } catch (error) {
          e = error;
          output.warn(`Error parsing ${recovPath}: ${e}: proceeding anyways`);
          return doHeal(cfgData, daemon, recovPath);
        }
        now = new Date().getTime();
        if (!(_.isNumber(recovData != null ? recovData.lastAttempt : void 0))) {
          output.warn(`last heal attempt ${recovData != null ? recovData.lastAttempt : void 0} is not a number: proceeding anyways`);
          return doHeal(cfgData, daemon, recovPath);
        }
        if (now - recovData.lastAttempt < 600000) { // 10 minutes
          return output.log(`last heal attempt ${recovData.lastAttempt} not too long ago: skipping.`);
        } else {
          return doHeal(cfgData, daemon, recovPath, recovData);
        }
      });
    });
  };

  // We've read the recovery data, and we're good to run
  doHeal = function(cfgData, daemon, recovPath, recovData) {
    var attempts, now, output, ref, str;
    output = daemon.output;
    output.log("Healing connection.");
    if (recovData != null) {
      now = new Date().getTime();
      attempts = ((ref = recovData.attempts) != null ? ref : 0) + 1;
      if (now - recovData.lastAttempt < 3600000) { // 1 hour reset
        attempts = 1;
      }
      recovData.lastAttempt = new Date().getTime();
      recovData.attempts = attempts;
    } else {
      recovData = {
        lastAttempt: new Date().getTime(),
        attempts: 1
      };
    }
    str = JSON.stringify(recovData, null, 2) + '\n';
    return fs.writeFile(recovPath, str, {
      encoding: 'utf8'
    }, function(err) {
      var singlePostUnload, singleton;
      if (err != null) {
        return output.warn(`can't write recovery data to ${recovPath}: ${err}: aborting`);
      }
      singleton = false;
      singlePostUnload = function(fromTimeout) {
        return function() {
          if (singleton) {
            return;
          }
          singleton = true;
          if (fromTimeout) {
            output.log("daemon unload timed out: proceeding anyways");
          }
          return postUnload(cfgData, daemon, recovData);
        };
      };
      // unloading will halt retries and heartbeat / loopback
      daemon.unloadNoUnlock(singlePostUnload(false));
      // give the daemon 10 second to unload, then proceed anyway
      return setTimeout(singlePostUnload(true), 10000);
    });
  };

  // We've unloaded the daemon (hopefully), and ready to save existing
  // registered tenants.
  postUnload = function(cfgData, daemon, recovData) {
    var output, savePath;
    output = daemon.output;
    output.log("Healer saving existing tenants.");
    // same savePath as in petit-clientd-bin.coffee
    savePath = path.join(path.dirname(cfgData.configPath), 'restore');
    return config.registeredTenants(cfgData, function(err, tenPath, dirs) {
      var str;
      str = JSON.stringify(dirs, null, 2) + '\n';
      return fs.writeFile(savePath, str, {
        encoding: 'utf8'
      }, function(err) {
        if (err != null) {
          output.warn(`couldn't write 'restore' file: ${err}`);
        }
        return postSaved(cfgData, daemon, recovData);
      });
    });
  };

  // We've saved existing registered tenants, let's unregister them
  postSaved = function(cfgData, daemon, recovData) {
    var handler, opts, output, ref, singlePostUnregister, singleton;
    output = daemon.output;
    output.log("Healer unregistering tenants.");
    singleton = false;
    singlePostUnregister = function(fromTimeout) {
      return function() {
        if (singleton) {
          return;
        }
        singleton = true;
        if (fromTimeout) {
          output.log("Healer unregister timed out: proceeding anyways");
        }
        return postUnregister(cfgData, daemon, recovData);
      };
    };
    // use the manager to try to unregister all tenants in their own backchannel
    opts = {
      options: {
        config: cfgData.configPath,
        all: true
      }
    };
    handler = new events.EventEmitter();
    handler.on('success', function(obj) {
      return singlePostUnregister(false)();
    });
    handler.on('failure', function(obj) {
      switch (obj.type) {
        case 'argv':
          output.warn(`Bad manager usage: ${obj.error}.`);
          break;
        case 'config':
          output.warn(`Couldn't read config \"${obj.path}\".`);
          break;
        case 'exists':
          output.warn("Client isn't registered.");
          break;
        case 'message':
          output.warn(`Couldn't unregister ${obj.path}.`);
          break;
        default:
          output.warn(`Unknown unregister error ${obj}`);
      }
      output.warn("Proceeding anyways.");
      return singlePostUnregister(false)();
    });
    manage.commands(handler).unregister.run(opts);
    // give unregistration 7 seconds per registered tenant
    return setTimeout(singlePostUnregister(true), 5000 + 7000 * ((ref = cfgData.registration) != null ? ref : []).length);
  };

  // We've unregistered our tenants, and want to kill other node processes
  postUnregister = function(cfgData, daemon, recovData) {
    var output;
    output = daemon.output;
    output.log(`Healer ${process.pid} killing other node clients.`);
    return psnode.lookup({
      command: 'node',
      arguments: 'clientd'
    }, function(err, list) {
      return async.eachSeries(list, function(proc, cb) {
        if (`${proc.pid}` !== `${process.pid}`) {
          output.log(`Healer killing ${proc.pid}`);
          return psnode.kill(proc.pid, function(err) {
            if (err != null) {
              output.warn(`Healer can't kill process ${proc.pid}: ${err}`);
            }
            return cb(null);
          });
        } else {
          return cb(null);
        }
      }, function(err) {
        if (err != null) {
          output.warn(`Healer problem killing processes: ${err}`);
        }
        return postKill(cfgData, daemon, recovData);
      });
    });
  };

  // We've killed other node processes, and are ready to delete / edit our
  // state files (so that the next startup will be relatively fresh). We'll
  // delete all the state files in the configPath, and all of them in each
  // tenant as well. Unless this is a later attempt, we won't delete the
  // "preferred" file: instead, we'll just remove the "port" property,
  // which should let us keep our registered ID.
  postKill = function(cfgData, daemon, recovData) {
    var output;
    output = daemon.output;
    output.log("Healer editing / deleting files");
    return daemon.unlock(function() {
      var cfgDir, deleteFiles, names;
      cfgDir = path.dirname(cfgData.configPath);
      names = ['client.lock', 'client.pid', 'key', 'registration', 'endpoint'];
      deleteFiles = function() {
        return async.eachSeries(names, function(name, cb) {
          var fpath;
          fpath = path.join(cfgDir, name);
          return fs.exists(fpath, function(exists) {
            if (!exists) {
              return cb(null);
            }
            output.log(`Healer deleting file ${fpath}`);
            return fs.unlink(fpath, function(err) {
              if (err != null) {
                output.warn(`healer cannot remove ${fpath}: ${err}`);
              }
              return cb(null);
            });
          });
        }, function(err) {
          var doAllTens;
          if (err != null) {
            output.warn(`healer error ${err}`);
          }
          doAllTens = function() {
            return config.allTenants(cfgData, function(err, tenPath, tens) {
              if (err != null) {
                output.warn(`can't access tenants ${err}`);
                postDelete(cfgData, daemon, recovData);
              }
              return async.eachSeries(tens, function(ten, cb) {
                return async.eachSeries(names, function(name, cb) {
                  var fpath;
                  fpath = path.join(ten.dir, name);
                  return fs.exists(fpath, function(exists) {
                    if (!exists) {
                      return cb(null);
                    }
                    output.log(`Healer deleting file ${fpath}`);
                    return fs.unlink(fpath, function(err) {
                      if (err != null) {
                        output.warn(`healer can't remove ${fpath}: ${err}`);
                      }
                      return cb(null);
                    });
                  });
                }, function(err) {
                  if (err != null) {
                    output.warn(`healer error ${err}`);
                  }
                  if (names.indexOf('preferred') === -1) {
                    return editPreferred(output, path.join(ten.dir, 'preferred'), cb);
                  } else {
                    return cb(err);
                  }
                });
              }, function(err) {
                if (err != null) {
                  output.warn(`healer tenant error ${err}`);
                }
                return postDelete(cfgData, daemon, recovData);
              });
            });
          };
          if (names.indexOf('preferred') === -1) {
            return editPreferred(output, path.join(cfgDir, 'preferred'), function(err) {
              if (err != null) {
                output.warn(`healer preferred err ${err}`);
              }
              return doAllTens();
            });
          } else {
            return doAllTens();
          }
        });
      };
      // if recovData.attempts >= regIdResetTry then names.push 'preferred'
      return deleteFiles();
    });
  };

  // Edits a "preferred" file by deleting the port
  editPreferred = function(output, prefPath, cb) {
    output.log(`Healer editing preferred file ${prefPath}`);
    return fs.exists(prefPath, function(exists) {
      if (!exists) {
        return cb(null);
      }
      return fs.readFile(prefPath, {
        encoding: 'utf8'
      }, function(err, data) {
        var e, obj, str;
        if (err != null) {
          return cb(err);
        }
        try {
          obj = JSON.parse(data);
        } catch (error) {
          e = error;
          return cb(e);
        }
        delete obj.port;
        str = JSON.stringify(obj, null, 2) + '\n';
        return fs.writeFile(prefPath, str, {
          encoding: 'utf8'
        }, cb);
      });
    });
  };

  // We've done everything: now we'll restart.
  postDelete = function(cfgData, daemon, recovData) {
    var output;
    output = daemon.output;
    output.log("Healer restarting.");
    util.restart(daemon.service);
    // We should have killed off all our threads by this point.
    // However, just in case:
    return process.exit(-1);
  };

}).call(this);
