(function() {
  // manage.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A simple top-level management program to observe and control a running
  // client daemon.
  var _, async, backtalk, childs, commands, config, events, forge, fs, lock, net, path, rimraf, util;

  _ = require('lodash');

  async = require('async');

  childs = require('child_process');

  events = require('events');

  forge = require('node-forge');

  fs = require('fs');

  net = require('net');

  path = require('path');

  rimraf = require('rimraf');

  config = require('../config');

  lock = require('../lock');

  backtalk = require('./backtalk');

  util = require('./util');

  exports.commands = commands = function(emitter) {
    var evictRun, fail, multi, part, stop, succ, unregisterRun;
    succ = function(obj) {
      return emitter.emit('success', obj);
    };
    fail = function(obj) {
      return emitter.emit('failure', obj);
    };
    multi = function(m) {
      return emitter.emit('multi', m);
    };
    part = function(...p) {
      return emitter.emit('part', ...p);
    };
    return {
      init: {
        opts: [],
        run: function(opts) {
          return config.createConfig(opts.options.config, function(err, made, path) {
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: path
              });
            }
            if (made) {
              return succ({
                path: path
              });
            } else {
              return fail({
                type: 'exists',
                path: path
              });
            }
          });
        }
      },
      start: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, configPath, data) {
            var exe, noder;
            if (err != null) {
              return fail({
                type: 'config',
                path: configPath,
                error: err
              });
            }
            exe = path.join(path.dirname(fs.realpathSync(__filename)), 'petit-clientd-bin') + path.extname(__filename);
            noder = path.extname(exe) === '.coffee' ? 'coffee' : 'node';
            return lock.findDeep('client', configPath, true, function(err, exists, data) {
              var ch, checkgo, totalTime;
              if (err != null) {
                return fail({
                  type: 'lock',
                  error: err
                });
              }
              if (exists) {
                return fail({
                  type: 'exists',
                  pid: data.pid
                });
              }
              ch = childs.spawn(noder, [exe, '-c', configPath], {
                detached: true,
                stdio: 'ignore'
              });
              ch.once('error', function(err) {
                return fail({
                  type: 'message',
                  message: "Unable to start client.",
                  error: err
                });
              });
              ch.unref();
              totalTime = 0;
              checkgo = function() {
                return lock.find('client', configPath, function(err, exists, data) {
                  if (err != null) {
                    return fail({
                      type: 'postlock',
                      error: err
                    });
                  }
                  if (exists) {
                    return succ({
                      pid: data.pid
                    });
                  } else if (totalTime > 4000) {
                    return fail({
                      type: 'postfind'
                    });
                  } else {
                    totalTime += 200;
                    return setTimeout(checkgo, 200);
                  }
                });
              };
              return setTimeout(checkgo, 200);
            });
          });
        }
      },
      stop: stop = {
        opts: [],
        run: function(opts) {
          var configPath, isKilled, waitKilled;
          configPath = config.pickConfig(opts.options.config);
          lock.find('client', configPath, function(err, exists, data) {
            var conn;
            if (err != null) {
              return fail({
                type: 'lock',
                error: err
              });
            }
            if (!exists) {
              return fail({
                type: 'exists'
              });
            }
            conn = new net.Socket();
            conn.setEncoding('utf8');
            conn.on('close', function() {
              return waitKilled(data.pid, function(killed) {
                if (killed) {
                  return succ({
                    sigint: false
                  });
                }
                process.kill(data.pid, 'SIGINT');
                return waitKilled(data.pid, function(killed) {
                  if (killed) {
                    return succ({
                      sigint: true
                    });
                  } else {
                    return fail({
                      type: 'unstop',
                      pid: data.pid
                    });
                  }
                });
              });
            });
            return conn.connect(data.listen, 'localhost', function() {
              return conn.end('STOP\x03');
            });
          });
          waitKilled = function(pid, cb) {
            var maybeDone, repeater, total;
            if (isKilled(pid)) {
              return cb(true);
            }
            total = 0;
            maybeDone = function() {
              if (isKilled(pid)) {
                clearInterval(repeater);
                return cb(true);
              } else if (total < 10) {
                return ++total;
              } else {
                clearInterval(repeater);
                return cb(false);
              }
            };
            return repeater = setInterval(maybeDone, 100);
          };
          return isKilled = function(pid) {
            var e;
            try {
              process.kill(pid, 0);
            } catch (error) {
              e = error;
              return true;
            }
            return false;
          };
        }
      },
      status: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return lock.find('client', configPath, function(err, exists, piddata) {
            var conn;
            if (err != null) {
              return fail({
                type: 'lock',
                error: err
              });
            }
            if (!exists) {
              return fail({
                type: 'exists'
              });
            }
            conn = new net.Socket();
            conn.setEncoding('utf8');
            conn.on('data', function(data) {
              var obj;
              obj = JSON.parse(data);
              return succ({
                pid: piddata.pid,
                status: obj
              });
            });
            return conn.connect(piddata.listen, 'localhost', function() {
              return conn.end('STATUS\x03');
            });
          });
        }
      },
      sping: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return lock.find('client', configPath, function(err, exists, piddata) {
            var conn;
            if (err != null) {
              return fail({
                type: 'lock',
                error: err
              });
            }
            if (!exists) {
              return fail({
                type: 'exists'
              });
            }
            conn = new net.Socket();
            conn.setEncoding('utf8');
            conn.on('data', function(data) {
              var obj;
              obj = JSON.parse(data);
              succ(obj);
              return conn.end();
            });
            return conn.connect(piddata.listen, 'localhost', function() {
              return conn.write('PING\x03');
            });
          });
        }
      },
      log: {
        opts: [],
        run: function(opts) {
          var configPath;
          console.log("argv", opts.argv[0], opts.argv[1]);
          if (opts.argv.length < 2) {
            return fail({
              type: 'argv',
              error: "No arguments"
            });
          }
          configPath = config.pickConfig(opts.options.config);
          return lock.find('client', configPath, function(err, exists, piddata) {
            var conn;
            if (err != null) {
              return fail({
                type: 'lock',
                error: err
              });
            }
            if (!exists) {
              return fail({
                type: 'exists'
              });
            }
            conn = new net.Socket();
            conn.setEncoding('utf8');
            conn.on('data', function(data) {
              var obj;
              obj = JSON.parse(data);
              succ(obj);
              return conn.end();
            });
            return conn.connect(piddata.listen, 'localhost', function() {
              return conn.write(`LOG ${opts.argv[1]}\x03`);
            });
          });
        }
      },
      check: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, path, data) {
            var bt;
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: path
              });
            }
            bt = backtalk.create();
            bt.on('error', function(err) {
              return fail({
                type: 'message',
                message: "Unable to check.",
                error: err
              });
            });
            bt.on('connected', function() {
              return bt.check(function(err, data) {
                if (err != null) {
                  return fail(type('message', {
                    error: err
                  }));
                }
                succ(data);
                return bt.end();
              });
            });
            return bt.open(data.client);
          });
        }
      },
      upgrade: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, path, piddata) {
            var bt, dcb;
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: path
              });
            }
            bt = backtalk.create();
            bt.on('error', function(err) {
              return fail({
                type: 'message',
                message: "Unable to upgrade.",
                error: err
              });
            });
            dcb = function(err, data) {
              if (err != null) {
                return fail({
                  type: 'upgrade',
                  error: err
                });
              }
              part({
                type: 'downloaded',
                data: data
              });
              part({
                type: 'phase',
                verb: 'applying'
              });
              return lock.find('client', configPath, function(err, exists, cdata) {
                var buffer, conn;
                if (err != null) {
                  return rimraf(data.path, function(rerr) {
                    if (rerr != null) {
                      part({
                        type: 'error',
                        error: rerr
                      });
                    }
                    return fail({
                      type: 'lock',
                      error: err
                    });
                  });
                } else if (!exists) {
                  return rimraf(data.path, function(rerr) {
                    if (rerr != null) {
                      part({
                        type: 'error',
                        error: rerr
                      });
                    }
                    return fail({
                      type: 'exists'
                    });
                  });
                } else if (data.patch != null) {
                  conn = new net.Socket();
                  conn.setEncoding('utf8');
                  buffer = '';
                  conn.on('data', function(data) {
                    return buffer += data;
                  });
                  conn.on('end', function() {
                    var checkgo, obj, totalTime;
                    conn.end();
                    part({
                      type: 'applied'
                    });
                    part({
                      type: 'phase',
                      verb: 'restarting'
                    });
                    obj = JSON.parse(buffer);
                    rimraf(data.path, function(rerr) {
                      if (rerr != null) {
                        return part({
                          type: 'error',
                          error: rerr
                        });
                      }
                    });
                    if (!obj.success) {
                      return fail({
                        type: 'delegate',
                        error: obj.error
                      });
                    } else {
                      totalTime = 0;
                      checkgo = function() {
                        return lock.find('client', configPath, function(err, exists, newcdata) {
                          if (err != null) {
                            return fail({
                              type: 'postlock',
                              error: err
                            });
                          }
                          if (exists && newcdata.pid !== cdata.pid) {
                            return succ({
                              pid: newcdata.pid
                            });
                          } else if (totalTime > 6000) {
                            return fail({
                              type: 'postfind'
                            });
                          } else {
                            totalTime += 200;
                            return setTimeout(checkgo, 200);
                          }
                        });
                      };
                      return setTimeout(checkgo, 200);
                    }
                  });
                  conn.connect(cdata.listen, 'localhost', function() {
                    return conn.write(`UPGRADE ${JSON.stringify(data)}\x03`);
                  });
                } else {
                  succ(null);
                }
                return bt.end();
              });
            };
            bt.on('connected', function() {
              part({
                type: 'phase',
                verb: 'downloading'
              });
              return bt.upgrade(dcb, function(sofar, tot) {
                return part({
                  type: 'downloadStatus',
                  sofar: sofar,
                  total: tot
                });
              });
            });
            return bt.open(piddata.client);
          });
        }
      },
      reload: {
        opts: [],
        run: function(opts) {
          var configPath, isReload, waitReload;
          configPath = config.pickConfig(opts.options.config);
          lock.find('client', configPath, function(err, exists, piddata) {
            var conn;
            if (err != null) {
              return fail({
                type: 'lock',
                error: err
              });
            }
            if (!exists) {
              return fail({
                type: 'exists'
              });
            }
            conn = new net.Socket();
            conn.setEncoding('utf8');
            conn.on('close', function() {
              return waitReload(function(err, rel) {
                if (err != null) {
                  return fail({
                    type: 'error',
                    error: err
                  });
                }
                if (!rel) {
                  return fail({
                    type: 'reload'
                  });
                }
                return succ({});
              });
            });
            return conn.connect(piddata.listen, 'localhost', function() {
              return conn.end('RELOAD\x03');
            });
          });
          waitReload = function(cb) {
            return isReload(function(err, rel) {
              var maybeDone, repeater, total;
              if (err != null) {
                return cb(err);
              }
              if (rel) {
                return cb(null, true);
              }
              maybeDone = function() {
                return isReload(function(err, rel) {
                  if (err != null) {
                    clearInterval(repeater);
                    return cb(err);
                  } else if (rel) {
                    clearInterval(repeater);
                    return cb(null, true);
                  } else if (total < 60) {
                    return ++total;
                  } else {
                    clearInterval(repeater);
                    return cb(null, false);
                  }
                });
              };
              repeater = setInterval(maybeDone, 100);
              return total = 0;
            });
          };
          return isReload = function(cb) {
            return lock.find('client', configPath, cb);
          };
        }
      },
      config: {
        opts: [['t', 'tenant=ARG', 'find tenant config']],
        run: function(opts) {
          return config.readConfig(opts.options.config, function(err, configPath, configData) {
            if (err != null) {
              return fail({
                type: 'config',
                path: configPath,
                error: err
              });
            }
            if (!opts.options.tenant) {
              return util.verifyConfig(configPath, configData, function(err, isReg) {
                var fingerprint, keydata, privateKey, registration;
                if (isReg) {
                  if (err) {
                    return fail({
                      type: 'verify',
                      error: err
                    });
                  }
                  registration = configData.client.registration;
                  delete configData.client.registration;
                  keydata = configData.client.keydata;
                  delete configData.client.keydata;
                  privateKey = forge.pki.privateKeyFromPem(keydata);
                  fingerprint = forge.ssh.getPublicKeyFingerprint(privateKey, {
                    encoding: 'hex',
                    delimiter: ':'
                  });
                  multi([
                    [
                      "Registered with IDs",
                      _.map(registration,
                      function(r) {
                        return r.id;
                      })
                    ],
                    ["Key fingerprint",
                    fingerprint],
                    [`Verified configuration file at ${configPath}.`],
                    ["\"config\":",
                    JSON.stringify(configData.client,
                    null,
                    2)]
                  ]);
                  return succ({});
                } else {
                  multi([["This client isn't registered."], [`Verified configuration file at ${configPath}.`], ["\"config\":", JSON.stringify(configData.client, null, 2)]]);
                  return succ({});
                }
              });
            } else {
              return config.tenantConfig(opts.options.tenant, configData, function(err, tCfgPath, tCfgData) {
                if (err) {
                  return fail({
                    type: 'tenant',
                    error: err
                  });
                }
                multi([[`Tenant config at ${tCfgPath}.`], ["\"config\":", JSON.stringify(tCfgData, null, 2)]]);
                return succ({});
              });
            }
          });
        }
      },
      lease: {
        opts: [['l', 'host=ARG', 'set the local service host'], ['p', 'port=ARG', 'set the local service port'], ['s', 'secure=ARG', '(true|false) if the local service is tls']],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, configPath, configData) {
            var e, ref;
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: configPath
              });
            }
            if ((opts.options.secure != null) && opts.options.secure === 'true') {
              _.set(configData, 'client.local.tls', true);
            } else {
              _.set(configData, 'client.local.tls', false);
            }
            if (opts.options.host != null) {
              _.set(configData, 'client.local.host', opts.options.host);
            }
            if (opts.options.port != null) {
              try {
                _.set(configData, 'client.local.port', parseInt(opts.options.port));
              } catch (error) {
                e = error;
              }
            }
            return config.lease((ref = configData.client) != null ? ref : {}, function(err, leasePath) {
              if (err != null) {
                return fail({
                  type: 'lease',
                  error: err,
                  path: leasePath
                });
              }
              return succ({
                path: leasePath
              });
            });
          });
        }
      },
      register: {
        opts: [['t', 'tenant=ARG', 'register by tenant']],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, configPath, configData) {
            var bt;
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: configPath
              });
            }
            bt = backtalk.create();
            bt.on('error', function(err) {
              return fail({
                type: 'message',
                message: "Unable to register.",
                error: err
              });
            });
            bt.on('connected', function() {
              return bt.register(opts.options.tenant);
            });
            bt.on('registered', function(regPath) {
              succ({
                path: regPath
              });
              return bt.end();
            });
            bt.on('keygen start', function() {
              return part("Generating key (this may take a few minutes) ...");
            });
            bt.on('keygen end', function() {
              return part("Key generated.");
            });
            return bt.open(configData.client);
          });
        }
      },
      unregister: {
        opts: [['r', 'regid=ARG', 'unregister by regid'], ['t', 'tenant=ARG', 'unregister by tenant'], ['a', 'all', 'unregister all']],
        run: unregisterRun = function(opts) {
          var configPath, optsuse;
          optsuse = 0;
          if (opts.options.regid) {
            optsuse += 1;
          }
          if (opts.options.tenant) {
            optsuse += 1;
          }
          if (opts.options.all) {
            optsuse += 1;
          }
          if (optsuse !== 1) {
            return fail({
              type: 'argv',
              error: "Must use exactly one of -a -r -t."
            });
          }
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, configPath, data) {
            var bt, tenPath;
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: configPath
              });
            }
            if (opts.options.regid) {
              bt = backtalk.create();
              bt.on('error', function(err) {
                return fail({
                  type: 'message',
                  message: 'Unable to unregister.',
                  error: err
                });
              });
              bt.on('connected', function() {
                return bt.unregister(opts.options.regid);
              });
              bt.on('unregistered', function(regPath) {
                succ({
                  path: regPath
                });
                return bt.end();
              });
              return bt.open(data.client);
            } else if (opts.options.tenant) {
              tenPath = path.join(path.dirname(configPath), 'tenants', opts.options.tenant);
              return fs.exists(tenPath, function(exists) {
                if (!exists) {
                  return fail({
                    type: 'unregister',
                    error: `No such tenant ${tenPath}.`,
                    path: tenPath
                  });
                } else {
                  return fs.exists(path.join(tenPath, 'registration'), function(exists) {
                    if (exists) {
                      return fs.readFile(path.join(tenPath, 'registration'), {
                        encoding: 'utf8'
                      }, function(err, data) {
                        var e, obj, opts2;
                        try {
                          obj = JSON.parse(data.trim());
                        } catch (error) {
                          e = error;
                          return fail({
                            type: 'unregister',
                            error: e
                          });
                        }
                        opts2 = _.cloneDeep(opts);
                        delete opts2.options.tenant;
                        opts2.options.regid = obj.id;
                        return unregisterRun(opts2);
                      });
                    } else {
                      return fail({
                        type: 'unregister',
                        error: `Tenant ${tenPath} is not registered.`,
                        path: tenPath
                      });
                    }
                  });
                }
              });
            } else if (opts.options.all) {
              return config.registeredTenants(data, function(err, tenPath, tens) {
                if (err != null) {
                  return fail({
                    type: 'error',
                    error: err
                  });
                }
                return async.mapSeries(tens, function(ten, cb) {
                  var em2, opts2;
                  em2 = new events.EventEmitter();
                  em2.on('success', function(obj) {
                    return cb(null, obj.path);
                  });
                  em2.on('failure', function(obj) {
                    return cb(obj);
                  });
                  em2.on('multi', function(m) {
                    return multi(m);
                  });
                  em2.on('part', function(...p) {
                    return part(...p);
                  });
                  opts2 = _.cloneDeep(opts);
                  delete opts2.options.all;
                  opts2.options.tenant = ten;
                  return commands(em2).unregister.run(opts2);
                }, function(err, paths) {
                  if (err != null) {
                    return fail(err);
                  }
                  return succ({
                    paths: paths
                  });
                });
              });
            } else {
              return fail({
                type: 'argv',
                error: "Must use exactly one of -a -r -t."
              });
            }
          });
        }
      },
      evict: {
        opts: [['r', 'regid=ARG', 'evict by regid'], ['t', 'tenant=ARG', 'evict by tenant']],
        run: evictRun = function(opts) {
          var configPath;
          if (!opts.options.regid && !opts.options.tenant) {
            return fail({
              type: 'argv',
              error: "No target specified (use '-r' or '-t')."
            });
          }
          if (opts.options.regid && opts.options.tenant) {
            return fail({
              type: 'argv',
              error: "Can't use '-r' and '-t' together."
            });
          }
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, configPath, data) {
            var bt, tenPath;
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: configPath
              });
            }
            if (opts.options.regid) {
              bt = backtalk.create();
              bt.on('error', function(err) {
                return fail({
                  type: 'message',
                  message: 'Unable to evict.',
                  error: err
                });
              });
              bt.on('connected', function() {
                return bt.unregister(opts.options.regid);
              });
              bt.on('unregistered', function(regPath) {
                var tenPath;
                tenPath = path.dirname(regPath);
                return rimraf(tenPath, function(err) {
                  if (err != null) {
                    return fail({
                      type: 'evict',
                      error: err,
                      path: tenPath
                    });
                  }
                  succ({
                    path: tenPath
                  });
                  return bt.end();
                });
              });
              return bt.open(data.client);
            } else if (opts.options.tenant) {
              tenPath = path.join(path.dirname(configPath), 'tenants', opts.options.tenant);
              return fs.exists(tenPath, function(exists) {
                if (!exists) {
                  return fail({
                    type: 'evict',
                    error: `No such tenant ${tenPath}.`,
                    path: tenPath
                  });
                } else {
                  return fs.exists(path.join(tenPath, 'registration'), function(exists) {
                    if (exists) {
                      return fs.readFile(path.join(tenPath, 'registration'), {
                        encoding: 'utf8'
                      }, function(err, data) {
                        var e, obj;
                        try {
                          obj = JSON.parse(data.trim());
                        } catch (error) {
                          e = error;
                          return fail({
                            type: 'evict',
                            error: e
                          });
                        }
                        return evictRun({
                          options: {
                            regid: obj.id
                          }
                        });
                      });
                    } else {
                      return rimraf(tenPath, function(err) {
                        if (err != null) {
                          return fail({
                            type: 'message',
                            error: `Unable to evict ${tenPath}: ${err}`
                          });
                        }
                        return succ({
                          path: tenPath
                        });
                      });
                    }
                  });
                }
              });
            }
          });
        }
      },
      tenants: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, configPath, configData) {
            if (err != null) {
              return fail({
                type: 'config',
                error: err,
                path: configPath
              });
            }
            return config.allTenants(configData, function(err, tenDir, tens) {
              if (err != null) {
                return fail({
                  type: 'tenants',
                  error: err
                });
              }
              return async.each(tens, function(ten, cb) {
                return config.tenantConfig(ten.name, configData, function(err, tCfgPath, tCfgData) {
                  ten.config = tCfgData;
                  return cb(null);
                });
              }, function(err) {
                return succ({
                  tenants: tens
                });
              });
            });
          });
        }
      },
      ping: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return lock.find('client', configPath, function(err, exists, data) {
            if (err != null) {
              return fail({
                type: 'lock',
                error: err
              });
            }
            if (exists) {
              return fail({
                type: 'exists',
                pid: data.pid
              });
            }
            return config.readConfig(configPath, function(err, configPath, configData) {
              if (err != null) {
                return fail({
                  type: 'config',
                  error: err,
                  path: configPath
                });
              }
              return util.verifyConfig(configPath, configData, function(err, isReg) {
                var bt;
                if (err) {
                  return fail({
                    type: 'verify',
                    error: err
                  });
                }
                bt = backtalk.create();
                bt.on('error', function(err) {
                  return fail({
                    type: 'message',
                    message: 'Unable to ping.',
                    error: err
                  });
                });
                bt.on('connected', function() {
                  return part("Connected.");
                });
                bt.on('handshake', function() {
                  succ({});
                  return bt.end();
                });
                return bt.connect({
                  config: configData.client
                });
              });
            });
          });
        }
      },
      version: {
        opts: [],
        run: function(opts) {
          var configPath;
          configPath = config.pickConfig(opts.options.config);
          return config.readConfig(configPath, function(err, configPath, configData) {
            if (err != null) {
              return fail({
                message: 'config err',
                error: err
              });
            }
            return succ(configData.runtime || config.version());
          });
        }
      }
    };
  };

}).call(this);
