(function() {
  var applyDiff, applyDiffChunk, applyMain, async, diff, fs, mkdirp, path, resolvePath, targz, unlink, util;

  async = require('async');

  fs = require('fs');

  path = require('path');

  targz = require('tar');

  mkdirp = require('mkdirp');

  diff = require('diff');

  util = require('./util');

  exports.makeDest = function(config, cb) {
    var upath;
    if (config.configPath == null) {
      return cb(new Error("No configPath listed"));
    }
    upath = path.join(path.dirname(config.configPath), 'upgrades', util.guid());
    return mkdirp(upath).then((err) => {
      if (err != null) {
        return cb(new Error(`Couldn't make upgrades dir: ${err}`));
      }
      return cb(null, {
        path: function() {
          return upath;
        },
        outstream: function() {
          return targz().createWriteStream(upath);
        }
      });
    });
  };

  exports.applyPatch = function(patch, ppath, configPath, cb) {
    var main;
    if (!patch.main) {
      return fs.readdir(ppath, function(err, files) {
        var mains;
        mains = _.filter(files, function(f) {
          return f.startsWith('main.');
        });
        if (mains.length !== 1) {
          return cb(new Error(`Couldn't choose main from ${mains}.`));
        }
        return applyMain(patch, ppath, configPath, path.join(ppath, mains[0]), cb);
      });
    } else {
      main = path.join(ppath, patch.main);
      return fs.exists(main, function(exists) {
        if (exists) {
          return applyMain(patch, ppath, main, configPath, cb);
        } else {
          return cb(new Error(`Main file ${main} doesn't exist.`));
        }
      });
    }
  };

  exports.applyMain = applyMain = function(patch, ppath, main, configPath, cb) {
    switch (path.extname(main)) {
      case '.js':
        return applyJs(patch, ppath, main, configPath, cb);
      case '.exe':
        return applyExe(patch, ppath, main, configPath, cb);
      case '.diff':
      case '.patch':
        return applyDiff(patch, ppath, main, configPath, cb);
      default:
        return cb(new Error(`Unknown main type ${main}.`));
    }
  };

  exports.applyDiff = applyDiff = function(patch, ppath, main, configPath, cb) {
    var chunk, chunkStart, ended, fin, readLoop, reading;
    fin = fs.createReadStream(main, {
      encoding: 'utf8'
    });
    chunk = '';
    chunkStart = 0;
    reading = false;
    ended = false;
    readLoop = function(err) {
      var m, newchunk;
      if (err != null) {
        return cb(err);
      }
      reading = true;
      if ((m = chunk.match(/\n[^ \-\+\@\\]/)) != null) {
        newchunk = chunk.substr(0, m.index);
        chunkStart += m.index;
        chunk = chunk.substr(m.index + 1);
        return applyDiffChunk(newchunk, configPath, readLoop);
      } else if (ended) {
        newchunk = chunk;
        chunkStart += chunk.length;
        chunk = '';
        return applyDiffChunk(newchunk, configPath, cb);
      } else {
        return reading = false;
      }
    };
    return fin.on('readable', function() {
      var data;
      data = fin.read();
      if (data != null) {
        chunk += data;
      } else {
        ended = true;
      }
      if (!reading) {
        return readLoop();
      }
    });
  };

  applyDiffChunk = function(diffChunk, configPath, cb) {
    var dfile, dm, n, sfile, sm, withOrig;
    n = diffChunk.indexOf('\n');
    if (n >= 0) {
      diffChunk = diffChunk.substr(n + 1);
    }
    sm = /^--- ([^\n\r\t]+).*/m.exec(diffChunk);
    if (sm != null ? sm[1] : void 0) {
      sfile = resolvePath(sm[1], configPath);
    } else {
      return cb(new Error(`Unable to find source file in ${diffChunk}`));
    }
    dm = /^\+\+\+ ([^\n\r\t]+).*/m.exec(diffChunk);
    if (dm != null ? dm[1] : void 0) {
      dfile = resolvePath(dm[1], configPath);
    } else {
      return cb(new Error("Unable to find dest file"));
    }
    withOrig = function(err, orig) {
      if (err != null) {
        return cb(err);
      }
      return mkdirp(path.dirname(dfile)).then((err) => {
        var patched;
        if (err != null) {
          return cb(err);
        }
        patched = diff.applyPatch(orig, diffChunk);
        // git uses dest = '/dev/null' for file removals
        if (dfile === '/dev/null' && patched.length === 0) {
          return unlink(sfile, cb);
        } else if (patched.length === 0) {
          return unlink(dfile, cb);
        } else {
          return fs.writeFile(dfile, patched, {
            encoding: 'utf8'
          }, cb);
        }
      });
    };
    return fs.exists(sfile, function(exists) {
      if (exists) {
        return fs.readFile(sfile, {
          encoding: 'utf8'
        }, withOrig);
      } else {
        return withOrig(null, '');
      }
    });
  };

  unlink = function(fname, cb) {
    return fs.exists(fname, function(exists) {
      if (exists) {
        return fs.unlink(fname, cb);
      } else {
        return cb(null);
      }
    });
  };

  resolvePath = function(sfile, configPath) {
    // git uses 'a/' vs 'b/'
    if (sfile.startsWith('a/') || sfile.startsWith('b/')) {
      sfile = sfile.substr(2);
    }
    if (!path.parse(sfile).root) {
      sfile = path.join(path.dirname(configPath), sfile);
    }
    return sfile;
  };

}).call(this);
