#!/usr/bin/env node

(function() {
  //!/usr/bin/env coffee

  // petit-client-bin.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.
  var _, commands, events, getopt, main, outputs, say, stdEmit, warn, wordwrap;

  _ = require('lodash');

  getopt = require('node-getopt');

  events = require('events');

  wordwrap = require('wordwrap');

  say = console.log;

  warn = console.warn;

  commands = require('./manage').commands;

  main = function() {
    var cmd, opt, opts, optsrc, ref, ref1, task, usage;
    usage = function() {
      say("Usage: petit-client <command> [OPTIONS]");
      say("Known commands are:");
      say(wordwrap(2, 72)(_.sortBy(_.keys(commands())).join(', ')));
      return process.exit(1);
    };
    if (process.argv.length < 3 || ((ref = process.argv[2]) === '-h' || ref === '--help')) {
      usage();
    }
    cmd = process.argv[2];
    task = commands(stdEmit(cmd))[cmd];
    if (task == null) {
      say(`Unknown command \"${process.argv[2]}\"`);
      usage();
    }
    optsrc = (ref1 = task.opts) != null ? ref1 : [];
    optsrc.push(...[['c', 'config=ARG', 'use config file'], ['v', 'verbose', 'print debug output'], ['h', 'help', 'display this help']]);
    opt = getopt.create(optsrc);
    opt.setHelp(`Usage: petit-client ${process.argv[2]} [OPTION]

[[OPTIONS]]`);
    opts = opt.bindHelp().parseSystem();
    return task.run(opts);
  };

  stdEmit = function(cmd) {
    var emitter, output;
    emitter = new events.EventEmitter();
    output = outputs[cmd];
    if (typeof output === 'function') {
      output = output();
    }
    emitter.on('success', function(obj) {
      var out;
      if ((out = output != null ? output.succ : void 0) != null) {
        return out(obj);
      } else {
        return say(`Uncategorized ${cmd} success: ${JSON.stringify(obj)}.`);
      }
    });
    emitter.on('failure', function(obj) {
      var out;
      if ((out = output != null ? output.fail : void 0) != null) {
        out(obj);
      } else {
        warn(`Uncategorized ${cmd} failure: ${JSON.stringify(obj)}.`);
      }
      return process.exit(101);
    });
    emitter.on('multi', function(multi) {
      var i, len, m, results;
      results = [];
      for (i = 0, len = multi.length; i < len; i++) {
        m = multi[i];
        results.push(say(...m));
      }
      return results;
    });
    emitter.on('part', function(...part) {
      var out;
      if ((out = output != null ? output.part : void 0) != null) {
        return out(...part);
      } else {
        return say(...part);
      }
    });
    return emitter;
  };

  outputs = {
    init: {
      succ: function(obj) {
        return say(`Created configuration file \"${obj.path}\".`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'config':
            return warn(`Couldn't create file \"${obj.path}\".`, obj.error);
          case 'exists':
            return warn(`Configuration file \"${obj.path}\" exists.`);
          default:
            return warn("Unknown init error", obj.message, obj.error);
        }
      }
    },
    start: {
      succ: function(obj) {
        return say(`Client started on pid ${obj.pid}.`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'lock':
            return warn("Couldn't lock client", obj.error);
          case 'exists':
            return warn(`Client is already running on pid ${obj.pid}`);
          case 'message':
            return warn(obj.message, obj.error);
          case 'postlock':
            return warn("Couldn't lock after starting", obj.error);
          case 'postfind':
            return warn("Couldn't start client.");
          default:
            return warn("Unknown start error", obj.message, obj.error);
        }
      }
    },
    stop: {
      succ: function(obj) {
        if (obj.sigint) {
          return say("Client is stopped with SIGINT.");
        } else {
          return say("Client stopped.");
        }
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'lock':
            return warn("Couldn't lock client", obj.error);
          case 'exists':
            return warn("Client is not running.");
          case 'unstop':
            return warn(`Unable to stop process on pid ${obj.pid}.`);
          default:
            return warn("Unknown stop error", obj.message, obj.error);
        }
      }
    },
    status: {
      succ: function(obj) {
        return say(`Client has status on pid ${obj.pid}\n` + JSON.stringify(obj.status, null, 2));
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'lock':
            return warn("Couldn't lock client", obj.error);
          case 'exists':
            return warn("Client isn't running.");
          default:
            return warn("Unknown status error", obj.message, obj.error);
        }
      }
    },
    sping: {
      succ: function(obj) {
        return say(`Client ping: ${JSON.stringify(obj)}`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'lock':
            return warn("Couldn't lock client", obj.error);
          case 'exists':
            return warn("Client isn't running.");
          default:
            return warn("Unknown ping error", obj.message, obj.error);
        }
      }
    },
    log: {
      succ: function(obj) {
        return say(`Logging level set to \"${obj.level}\".`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'argv':
            return warn("Can't find log level to set to.");
          case 'lock':
            return warn("Couldn't find running client.");
          case 'exists':
            return warn("Client isn't running.");
          case 'log':
            return warn("Couldn't changing logging", obj.error);
          default:
            return warn("Unknown logset error", obj.message, obj.error);
        }
      }
    },
    check: {
      succ: function(obj) {
        return say(`Client check: ${JSON.stringify(obj)}`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'config':
            return warn("Couldn't read config", obj.error);
          case 'message':
            return warn("Message error.", obj.error);
          default:
            return warn("Unknown check error", obj.message, obj.error);
        }
      }
    },
    upgrade: function() {
      var lastDed;
      lastDed = 0;
      return {
        succ: function(obj) {
          return say(`Client upgraded and restarted on pid ${obj.pid}.`);
        },
        fail: function(obj) {
          switch (obj.type) {
            case 'config':
              return warn("Couldn't read config", obj.error);
            case 'message':
              return warn("Message error.", obj.error);
            case 'upgrade':
              return warn("Upgrade error.", obj.error);
            case 'exists':
              return warn("Client isn't running.");
            case 'error':
              return warn(`Couln't upgrade: ${obj.error}`);
            case 'lock':
              return warn("Couldn't lock client.", obj.error);
            case 'postlock':
              return warn("Couldn't lock client again.", obj.error);
            case 'postfind':
              return warn("Couldn't restart after upgrade.");
            case 'delegate':
              return warn("Failed delegated upgrade.", obj.error);
            case 'unstop':
              return warn(`Unable to stop process on pid ${obj.pid}.`);
            default:
              return warn("Unknown upgrade error", obj.type, obj.message, obj.error);
          }
        },
        part: function(part) {
          var results;
          switch (part.type) {
            case 'downloadStatus':
              if (part.sofar >= lastDed + 10000 || part.sofar === part.total) {
                say(`received ${part.sofar} / ${part.total} bytes`);
              }
              results = [];
              while (part.sofar >= lastDed + 10000) {
                lastDed = Math.min(lastDed + 10000, part.total);
                if (lastDed >= part.total) {
                  break;
                } else {
                  results.push(void 0);
                }
              }
              return results;
              break;
            case 'downloaded':
              return say('Download complete.', JSON.stringify(part.data));
            case 'applied':
              return say('Applied.');
            case 'phase':
              return say(`Now ${part.verb}...`);
            case 'error':
              return say(`Error ${part.error}`);
            default:
              return say("Unknown upgrade partial data", part.type);
          }
        }
      };
    },
    reload: {
      succ: function(obj) {
        return say("Client is reloaded.");
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'lock':
            return warn("Couldn't lock client", obj.error);
          case 'exists':
            return warn("Client isn't running.");
          case 'error':
            return warn("Error while reloading.", obj.error);
          case 'reload':
            return warn("Couldn't reload.");
          default:
            return warn("Unknown reload error", obj.message, obj.error);
        }
      }
    },
    config: {
      succ: function(obj) {},
      fail: function(obj) {
        switch (obj.type) {
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'verify':
            return warn("Couldn't verify configuration.", obj.error);
          case 'tenant':
            return warn("Couldn't get tenant.", obj.error);
          default:
            return warn("Unknown config error", obj.message, obj.error);
        }
      }
    },
    register: {
      succ: function(obj) {
        return say(`Registered to \"${obj.path}\".`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'exists':
            return warn("Client is already registered.");
          case 'message':
            return warn(obj.message, obj.error);
          default:
            return warn("Unknown register error", obj.message, obj.error);
        }
      }
    },
    lease: {
      succ: function(obj) {
        return say(`New lease at \"${obj.path}\".`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'lease':
            return warn(`Couldn't lease to ${obj.path}`);
          default:
            return warn("Unknown lease error", obj.message, obj.error);
        }
      }
    },
    unregister: {
      succ: function(obj) {
        if (obj.path != null) {
          return say(`Unregistered tenant ${obj.path}.`);
        } else {
          return say(`Unregistered all tenants ${obj.paths}.`);
        }
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'argv':
            return warn(`${obj.error}`);
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'exists':
            return warn("Client is not registered.");
          case 'message':
            return warn(obj.message, obj.error);
          case 'unregister':
            return warn(`Couldn't unregister tenant ${obj.path}`, obj.error);
          default:
            return warn("Unknown unregister error", obj.type, obj.message, obj.error);
        }
      }
    },
    evict: {
      succ: function(obj) {
        return say(`Evicted tenant at ${obj.path}.`);
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'argv':
            return warn(`${obj.error}`);
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'exists':
            return warn("Client is not registered.");
          case 'message':
            return warn(obj.message, obj.error);
          case 'evict':
            return warn(`Couldn't evict tenant ${obj.path}`, obj.error);
          case 'registered':
            return warn(`Already registered: ${obj.error}`);
          default:
            return warn("Unknown unregister error", obj.type, obj.message, obj.error);
        }
      }
    },
    tenants: {
      succ: function(obj) {
        var hostport, i, len, ref, ref1, ref2, ref3, ref4, results, secures, ten;
        if (_.isEmpty(obj.tenants)) {
          return say("No tenants found.");
        } else {
          say("Found tenants:");
          ref = obj.tenants;
          results = [];
          for (i = 0, len = ref.length; i < len; i++) {
            ten = ref[i];
            secures = ((ref1 = ten.config) != null ? (ref2 = ref1.local) != null ? ref2.tls : void 0 : void 0) ? "(tls)" : "(plain)";
            hostport = ((ref3 = ten.config) != null ? (ref4 = ref3.local) != null ? ref4.host : void 0 : void 0) ? `  ${ten.config.local.host}:${ten.config.local.port} ${secures}` : '';
            if (ten.registered) {
              results.push(say(` - ${ten.name} ${ten.registeredId}${hostport}`));
            } else {
              results.push(say(`   ${ten.name}${hostport}`));
            }
          }
          return results;
        }
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'tenants':
            return warn(`Couldn't find tenants ${obj.error}`);
          default:
            return warn("Unknown tenants error", obj.type, obj.message, obj.error);
        }
      }
    },
    ping: {
      succ: function(obj) {
        return say("Ping success.");
      },
      fail: function(obj) {
        switch (obj.type) {
          case 'lock':
            return warn("Couldn't lock client", obj.error);
          case 'exists':
            return warn("Client is running.");
          case 'config':
            return warn(`Couldn't read config \"${obj.path}\".`, obj.error);
          case 'verify':
            return warn("Couldn't verify configuration.", obj.error);
          case 'message':
            return warn(obj.message, obj.error);
          default:
            return warn("Unknown ping error", obj.message, obj.error);
        }
      }
    },
    version: {
      succ: function(obj) {
        var name, prov;
        name = obj.name || "Ground2Cloud Client";
        prov = obj.provider || "Cloud Elements <cloud-elements.com>";
        say(`${name} version ${obj.display}`);
        return say(`by ${prov}`);
      },
      fail: function(obj) {
        return warn("Unknown version error", obj.message, obj.error);
      }
    }
  };

  main();

}).call(this);
