#!/usr/bin/env node

(function() {
  //!/usr/bin/env coffee

  // petit-clientd-bin.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.
  var SEVERE, _, async, cfgPath, cl, client, config, daemon, errors, events, filelock, fs, getopt, handle, handleConfig, healer, heartbeat, heartintv, http, https, iceptPath, level, load, loading, lock, loopbeat, loopintv, main, manage, opt, output, patches, path, registerRestoreTenants, restapi, sendLoop, service, startHeartbeat, startRest, stopHeartbeat, tunnel, unload, unloadNoUnlock, unlock, url, util, winston;

  _ = require('lodash');

  async = require('async');

  client = require('./client');

  config = require('../config');

  errors = require('../errors');

  events = require('events');

  fs = require('fs');

  getopt = require('node-getopt');

  healer = require('./healer');

  http = require('http');

  https = require('https');

  lock = require('../lock');

  manage = require('./manage');

  patches = require('./patches');

  path = require('path');

  restapi = require('./restapi');

  tunnel = require('./tunnel');

  url = require('url');

  util = require('./util');

  winston = require('winston');

  SEVERE = errors.SEVERE;

  opt = getopt.create([['v', 'verbose', 'print debug output'], ['c', 'config=ARG', 'set config file'], ['s', 'service', 'this is a service run'], ['h', 'help', 'display this help']]).bindHelp().parseSystem();

  cfgPath = config.pickConfig(opt.options.config);

  level = opt.options.verbose ? 'verbose' : 'normal';

  service = opt.options.service;

  output = null;

  heartbeat = null;

  heartintv = 120000;

  loopbeat = null;

  loopintv = 300000;

  iceptPath = tunnel.iceptPath;

  daemon = {output, service};

  cl = client.create(daemon);

  filelock = null;

  main = function() {
    handle(cl);
    return load(function() {});
  };

  startHeartbeat = function(cb) {
    return stopHeartbeat(function() {
      var thump, whoosh;
      thump = function() {
        return cl.backchannel.ping(function(err, d) {
          if (err != null) {
            return output.warn(err);
          } else {
            return output.log(`Heartbeat success: ${d.type === 'didPing1'}`);
          }
        });
      };
      heartbeat = setInterval(thump, heartintv);
      whoosh = function() {
        return sendLoop(cl.config, function(err, d) {
          if (err != null) {
            return output.warn(`Loopback failure: ${err}`);
          } else {
            return output.log(`Loopback success: true for ${JSON.stringify(d)}`);
          }
        });
      };
      loopbeat = setInterval(whoosh, loopintv);
      return typeof cb === "function" ? cb() : void 0;
    });
  };

  stopHeartbeat = function(cb) {
    if (heartbeat != null) {
      clearInterval(heartbeat);
    }
    heartbeat = null;
    if (loopbeat != null) {
      clearInterval(loopbeat);
    }
    loopbeat = null;
    return typeof cb === "function" ? cb() : void 0;
  };

  sendLoop = function(config, cb) {
    var eps, ref, ref1;
    if ((eps = (ref = config.server) != null ? (ref1 = ref.vitae) != null ? ref1.endpoints : void 0 : void 0) == null) {
      return cb("No server endpoint found for loopback.");
    }
    return async.map(_.values(eps), function(ep, cb) {
      var caPath, certDir, handler, i, len, p, props, ref2;
      props = url.parse(ep);
      _.assign(props, {
        path: iceptPath,
        method: 'GET'
      });
      ref2 = ['host', 'href', 'pathname', 'search'];
      for (i = 0, len = ref2.length; i < len; i++) {
        p = ref2[i];
        delete props[p];
      }
      handler = function(res) {
        if (res.statusCode === 200) {
          return cb(null, ep);
        } else {
          return cb(`Loopback code: ${res.statusCode}`);
        }
      };
      switch (props.protocol) {
        case 'https:':
          if (props.port == null) {
            props.port = 443;
          }
          certDir = path.join(path.dirname(config.configPath), 'certificates');
          caPath = path.join(certDir, 'ca.bundle');
          if (fs.existsSync(caPath)) {
            props.ca = fs.readFileSync(caPath);
            props.agent = new https.Agent(props);
          }
          client = https.request(props, handler);
          break;
        case 'http:':
          if (props.port == null) {
            props.port = 80;
          }
          client = http.request(props, handler);
          break;
        default:
          return cb(`Unknown protocol ${props.protocol}`);
      }
      client.on('error', cb);
      return client.end();
    }, cb);
  };

  loading = false;

  daemon.load = load = function(cb) {
    if (loading) {
      return cb("Loading in progress.");
    }
    loading = true;
    output.log('Loading');
    return config.readConfig(opt.options.config, handleConfig(function(...args) {
      output.log('Load complete.');
      loading = false;
      return cb(...args);
    }));
  };

  daemon.unload = unload = function(cb) {
    if (loading) {
      return cb("Unloading in progress.");
    }
    loading = true;
    output.log('Unloading', filelock);
    if (filelock != null) {
      lock.unlock(filelock);
    }
    return stopHeartbeat(function() {
      output.log('Unload heartbeat stopped.');
      return cl.end(function() {
        output.log('Unload client ended.');
        return cl.deafen(function(...args) {
          output.log('Unload client deafened. Complete.');
          loading = false;
          return cb(...args);
        });
      });
    });
  };

  daemon.unloadNoUnlock = unloadNoUnlock = function(cb) {
    if (loading) {
      return cb("Unloading (w/out unlock) in progress.");
    }
    loading = true;
    return stopHeartbeat(function() {
      output.log('Unload heartbeat stopped.');
      return cl.end(function() {
        output.log('Unload client ended.');
        return cl.deafen(function(...args) {
          output.log('Unload client deafened. Complete.');
          loading = false;
          return cb(...args);
        });
      });
    });
  };

  daemon.unlock = unlock = function(cb) {
    if (filelock != null) {
      lock.unlock(filelock);
    }
    return cb(null);
  };

  handleConfig = function(cb) {
    return function(err, configPath, configData) {
      if (err != null) {
        output.prettyConfig(SEVERE, err);
        return cb(err);
      }
      // if there are tenants to register, register them before proceeding
      // qb connector itself should restart
      // qbmonitor.start configData.client
      return registerRestoreTenants(configData.client, function(err) {
        var ref;
        if (err != null) {
          output.warn("Couldn't restore some tenants");
        }
        if (((ref = configData.client) != null ? ref.logLevel : void 0) != null) {
          output.setLevel(configData.client.logLevel);
        }
        return util.verifyConfig(configPath, configData, function(err, isReg) {
          if (err != null) {
            output.prettyConfig(SEVERE, err);
            return cb(err);
          }
          return lock.lock('client', {
            pid: process.pid,
            listen: configData.client.listenPort
          }, configPath, configData, function(err, v) {
            var ref1;
            if (err != null) {
              output.warn(err);
              return cb(err);
            }
            filelock = v;
            if ((ref1 = configData.client) != null) {
              ref1.configPath = configData.configPath;
            }
            if (isReg) {
              return cl.listen(configData.client, function(err) {
                if (err != null) {
                  return cb(err);
                }
                return cl.start(configData.client, cb);
              });
            } else {
              return cl.listen(configData.client, cb);
            }
          });
        });
      });
    };
  };

  registerRestoreTenants = function(cfgData, cb) {
    var savePath;
    // same savePath as in healer.coffee
    savePath = path.join(path.dirname(cfgData.configPath), 'restore');
    return fs.exists(savePath, function(exists) {
      if (!exists) {
        return cb(null);
      }
      return fs.readFile(savePath, {
        encoding: 'utf8'
      }, function(err, data) {
        var e, tenantList;
        if (err != null) {
          output.warn(`Error reading ${savePath}: proceeding anyways`);
          fs.unlink(savePath);
          return cb(null);
        }
        try {
          tenantList = JSON.parse(data.trim());
        } catch (error) {
          e = error;
          output.warn(`Error parsing ${savePath}: ${e}: proceeding anyways`);
          fs.unlink(savePath);
          return cb(null);
        }
        if (!(_.isArray(tenantList))) {
          output.warn(`Data in ${savePath} is not an array`);
          fs.unlink(savePath);
          return cb(null);
        }
        return config.registeredTenants(cfgData, function(err, tenpath, dirs) {
          var newTenantList;
          if (err != null) {
            return cb(err);
          }
          newTenantList = [];
          return fs.unlink(savePath, function() {
            return async.eachSeries(tenantList, function(tenname, cb) {
              var handler, opts, singleCb, singleton;
              if ((dirs != null) && dirs.indexOf(tenname) !== -1) {
                return cb(null);
              }
              singleton = false;
              singleCb = function(fromTimeout, err) {
                return function() {
                  if (singleton) {
                    return;
                  }
                  singleton = true;
                  if (fromTimeout) {
                    output.log("Restore timed out: proceeding anyways");
                  }
                  if ((err != null) || fromTimeout) {
                    newTenantList.push(tenname);
                  }
                  return cb(null);
                };
              };
              opts = {
                options: {
                  config: cfgData.configPath,
                  tenant: tenname
                }
              };
              handler = new events.EventEmitter();
              handler.on('success', function(obj) {
                return singleCb(false)();
              });
              handler.on('failure', function(obj) {
                switch (obj.type) {
                  case 'config':
                    output.warn(`Can't read \"${obj.path}\".`);
                    break;
                  case 'exists':
                    output.warn("Client is already registered.");
                    break;
                  case 'message':
                    output.warn(obj.message, obj.error);
                    break;
                  default:
                    output.warn(`Unknown register error ${obj}`);
                }
                output.warn("Proceeding anyways.");
                return singleCb(false, "register failure")();
              });
              manage.commands(handler).register.run(opts);
              return setTimeout(singleCb(true), 5000);
            }, function(err) {
              var str;
              if (err != null) {
                return cb(err);
              }
              if (newTenantList.length > 0) {
                str = JSON.stringify(newTenantList, null, 2) + '\n';
                return fs.writeFile(savePath, str, {
                  encoding: 'utf8'
                }, function(err) {
                  if (err != null) {
                    output.warn(`couldn't write 'restore' file: ${err}`);
                  }
                  // if we can't register, then we probably can't talk to
                  // the server. Let's just wait 30 seconds and restart.
                  output.warn("Couldn't register some tenants during healing restart. Will attempt again in 30 seconds.");
                  return setTimeout((function() {
                    util.restart(daemon.service);
                    return process.exit(-1);
                  }), 30000);
                });
              } else {
                return cb(null);
              }
            });
          });
        });
      });
    });
  };

  handle = function(cl) {
    cl.on('error', function() {}); // handle in sub-structures
    cl.on('start', function() {
      return startHeartbeat(function() {
        return output.log('Client started.');
      });
    });
    cl.on('close', function(hadErr) {
      return stopHeartbeat(function() {
        return output.log('Client closed.');
      });
    });
    cl.on('listening', function() {
      return output.log('Client listening.');
    });
    cl.on('deaf', function() {
      return output.log('Client stopped listening.');
    });
    cl.on('retry', function(index) {
      return output.log('Retrying connection.');
    });
    cl.on('willRetry', function(index, timeout) {
      return output.log(`Connnection failed (${index}). Will retry in ${timeout} seconds.`);
    });
    cl.on('command', function(cmd, c) {
      var data, e, spind;
      output.log('Client command', cmd);
      switch (false) {
        case cmd !== 'STATUS':
          return c.end(cl.status());
        case cmd !== 'STOP':
          return stopHeartbeat(function() {
            return cl.end(function() {
              return cl.deafen();
            });
          });
        case cmd !== 'RELOAD':
          return unload(function() {
            return load(function() {});
          });
        case cmd !== 'PATCH':
          return unload;
        case cmd !== 'PING':
          return cl.backchannel.ping(function(err, stats) {
            if (err != null) {
              return c.end({
                type: 'error',
                error: `Can't ping: ${err}`
              });
            } else {
              return c.end(stats);
            }
          });
        case !_.startsWith(cmd, 'LOG'):
          spind = cmd.indexOf(' ');
          if (spind <= 0) {
            output.setLevel('debug');
            return c.end({
              success: true,
              level: 'debug'
            });
          } else {
            level = cmd.substr(spind + 1);
            output.setLevel(level);
            return c.end({
              success: true,
              level: level
            });
          }
          break;
        case !_.startsWith(cmd, 'UPGRADE'):
          spind = cmd.indexOf(' ');
          if (spind <= 0) {
            return c.end({
              success: false,
              error: "No path given for upgrade"
            });
          }
          try {
            data = JSON.parse(cmd.substr(spind + 1));
          } catch (error) {
            e = error;
            return c.end({
              success: false,
              error: e
            });
          }
          return patches.applyPatch(data.patch, data.path, cl.config.configPath, function(err) {
            if (err != null) {
              return c.end({
                success: false,
                error: `Patch failed: ${err}`
              });
            } else {
              c.end({
                success: true
              });
              stopHeartbeat(function() {
                return cl.end(function() {
                  return cl.deafen();
                });
              });
              return util.restart(service);
            }
          });
        default:
          return output.warn(`Unknown command ${cmd}`);
      }
    });
    cl.listener.on('error', function(err) {
      return output.prettyListen(SEVERE, err, cl.config);
    });
    cl.listener.on('close', function() {
      return output.log('Command listener closed.');
    });
    cl.listener.on('connection', function(c) {
      output.log('Command stream opened.');
      c.on('error', function(err) {
        return output.prettyListen(SEVERE, err, cl.config);
      });
      c.on('close', function() {
        return output.log('Command stream closed.');
      });
      return c.on('data', function(data) {
        return output.log('Command stream data', data);
      });
    });
    cl.listener.on('listening', function(addr) {
      return output.log('Command listening on', addr);
    });
    cl.backchannel.on('error', function(err) {
      return output.prettyBacktalk(SEVERE, err, cl.config);
    });
    cl.backchannel.on('close', function() {
      return output.log('Backchannel closed.');
    });
    cl.backchannel.on('connected', function() {
      return output.log('Backchannel connected.');
    });
    cl.backchannel.on('handshake', function() {
      return output.log('Backchannel handshake done.');
    });
    cl.backchannel.on('save', function(err, epath) {
      if (err != null) {
        return output.warn(`Couldn't save endpoint path ${epath}`, err);
      } else {
        return output.log(`Saved endpoint to ${epath}`);
      }
    });
    cl.tunnel.on('error', function(err) {
      output.prettyTunnel(SEVERE, err, cl.config);
      // this tunnel error could be indicative of a swiftpage-class problem
      // TODO: should we wait for a few of these to pile up before healing?
      return healer.heal(cl.config, daemon);
    });
    cl.tunnel.on('close', function(err) {
      return output.log('Tunnel closed.');
    });
    cl.tunnel.on('connected', function() {
      return output.log('Tunnel connected.');
    });
    return cl.tunnel.on('received', function(c) {
      output.log('Tunnel received.');
      c.on('bound', function() {
        return output.log('Local connection tunneled.');
      });
      c.on('unbound', function() {
        return output.log('Local connection closed.');
      });
      return c.on('error', function(err) {
        return output.prettyTunnel(SEVERE, err, cl.config);
      });
    });
  };

  startRest = function() {
    return config.readConfig(opt.options.config, function(err, cfgPath, cfgData) {
      var apiPort, ref, ref1;
      if (err != null) {
        return output.warn(`Can't start REST API: ${err}`);
      }
      apiPort = (ref = (ref1 = cfgData.client) != null ? ref1.apiPort : void 0) != null ? ref : 8100;
      output.log(`Starting REST API server on ${apiPort}`);
      return restapi.start(apiPort, opt, output);
    });
  };

  errors.init(cfgPath, 'client', level, function(err, out) {
    if (err != null) {
      return console.error('Unable to initialize logger', err);
    }
    daemon.output = output = out;
    startRest();
    return main();
  });

}).call(this);
