(function() {
  // qbmonitor.coffee    Copyright (c) 2017 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // The side-program that manages the COM-based "CE Connector"
  // application, which may be installed by the Ground2Cloud installer.
  var async, config, errors, exec, fs, log, mkdirp, path, readQbLocation, regedit, runQbDelayed, runQbForever, spawn, start, startLater, stop;

  async = require('async');

  ({exec, spawn} = require('child_process'));

  config = require('../config');

  path = require('path');

  fs = require('fs');

  errors = require('../errors');

  log = errors.create('info');

  mkdirp = require('mkdirp');

  regedit = require('regedit');

  exports.stop = stop = function() {};

  exports.start = start = function(configData) {
    return setTimeout((function() {
      return startLater(configData);
    }), 20000);
  };

  startLater = function(configData) {
    if (configData.qbmonitor == null) {
      configData.qbmonitor = {
        shouldRun: false
      };
    }
    return readQbLocation(configData, function(err, loc) {
      if (err != null) {
        log.log("Couldn't read external plugin.");
        return start(configData);
      } else {
        return runQbForever(configData, loc, function(err) {
          if (err != null) {
            log.warn(`QB Monitor exited prematurely: ${err}`);
          }
        });
      }
    });
  };

  readQbLocation = function(configData, cb) {
    return regedit.list('HKCU\\SOFTWARE\\CE', function(err, result) {
      var ifldr, ref, ref1, ref2;
      if (err != null) {
        return cb(err);
      }
      ifldr = (ref = result['HKCU\\SOFTWARE\\CE']) != null ? (ref1 = ref.values) != null ? (ref2 = ref1.InstallFolder) != null ? ref2.value : void 0 : void 0 : void 0;
      if (ifldr) {
        ifldr = ifldr + '\\QB Connector.exe';
      }
      if (ifldr != null) {
        return cb(null, ifldr);
      } else {
        return cb("No QB location could be found.");
      }
    });
  };

  runQbForever = function(configData, loc, cb) {
    return setTimeout((function() {
      return runQbDelayed(configData, loc, cb);
    }), 15000);
  };

  runQbDelayed = function(configData, loc, cb) {
    var args, cmd, err, g2cdir, logDirPath, out;
    cmd = loc;
    g2cdir = path.dirname(configData.configPath);
    // args = ['-port', "#{port}",
    //         '-app', 'CloudElements',
    //         '-g2cdir', "#{g2cdir}"
    // ];
    args = [];
    logDirPath = path.join(g2cdir, 'logs');
    out = 'ignore';
    err = 'ignore';
    return mkdirp(logDirPath).then((err) => {
      var child;
      if (err != null) {
        log.warn("Unable to make log dir for connector logs.");
      } else {
        out = fs.openSync(path.join(logDirPath, './ceconnector.log'), 'w');
        err = fs.openSync(path.join(logDirPath, './ceconnector.err'), 'w');
      }
      child = spawn(cmd, args, {
        detached: false,
        stdio: ['ignore', out, err]
      });
      // child.unref()
      // cb null
      return child.on('exit', function(code) {
        log.warn(`Child QB exited with code ${code}.`);
        return cb(null);
      });
    });
  };

}).call(this);
