(function() {
  // restapi.coffee    Copyright (c) 2017 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.
  var _, argOpts, cloneOpts, command, events, manage, responseHandler, sendTo;

  _ = require('lodash');

  events = require('events');

  manage = require('./manage');

  exports.start = function(port, opts, output) {
    var allowCrossDomain, app, bodyParser, cmd, express;
    express = require('express');
    bodyParser = require('body-parser');
    allowCrossDomain = function(req, res, next) {
      var ref, ref1;
      res.header('Access-Control-Allow-Credentials', true);
      res.header('Access-Control-Allow-Origin', typeof req.headers === "function" ? req.headers(typeof origin !== "undefined" && origin !== null ? origin : '*') : void 0);
      res.header('Access-Control-Allow-Methods', 'PUT,OPTIONS,GET,DELETE,POST');
      res.header('Access-Control-Allow-Headers', (ref = (ref1 = req.headers) != null ? ref1['Access-Control-Request-Headers'] : void 0) != null ? ref : 'x-requested-with,content-type,authorization,accept');
      res.header('Access-Control-Max-Age', 1728000);
      return next();
    };
    app = express();
    app.use(allowCrossDomain);
    app.use(bodyParser.json({
      limit: '500kb'
    }));
    app.set('etag', false);
    // put something in the body to let crossDomain happen
    app.options('*', function(req, res, next) {
      return res.send(200);
    });
    // commands
    cmd = command(opts, output);
    app.get('/status', cmd.status);
    app.get('/tenants', cmd.tenants);
    app.delete('/tenants/:id', cmd.evict);
    // $ curl -H 'content-type:application/json' \
    //       -X POST localhost:8100/tenants \
    //       --data-binary '{"host": "abcdefg", "port": 12345}'
    app.post('/tenants', cmd.lease);
    app.get('/ping', cmd.ping);
    app.get('/version', cmd.version);
    app.post('/register', cmd.registerAny);
    app.post('/register/:id', cmd.register);
    app.post('/unregister/:id', cmd.unregister);
    app.get('/config', cmd.config);
    app.get('/config/:id', cmd.configTen);
    app.post('/reload', cmd.reload);
    app.post('/upgrade', cmd.upgrade);
    app.get('/check', cmd.check);
    app.post('/log', cmd.log);
    app.post('/sping', cmd.sping);
    app.post('/stop', cmd.stop);
    app.post('/start', cmd.start);
    app.listen(port);
    return output.log('REST API started');
  };

  command = function(opts, output) {
    var commands, k, loopfn, v, workers;
    workers = {
      tenants: function(manager, req, res) {
        return manager.tenants.run(opts);
      },
      status: function(manager, req, res) {
        return manager.status.run(opts);
      },
      evict: function(manager, req, res, next) {
        return manager.evict.run(cloneOpts(opts, {
          tenant: req.params.id
        }));
      },
      registerAny: function(manager, req, res, next) {
        return manager.register.run(opts);
      },
      register: function(manager, req, res, next) {
        return manager.register.run(cloneOpts(opts, {
          tenant: req.params.id
        }));
      },
      unregister: function(manager, req, res, next) {
        return manager.unregister.run(cloneOpts(opts, {
          tenant: req.params.id
        }));
      },
      lease: function(manager, req, res, next) {
        var params;
        params = {};
        if (req.body.host != null) {
          params.host = req.body.host;
        }
        if (req.body.port != null) {
          params.port = req.body.port;
        }
        return manager.lease.run(cloneOpts(opts, params));
      },
      ping: function(manager, req, res) {
        return manager.ping.run(opts);
      },
      version: function(manager, req, res) {
        return manager.version.run(opts);
      },
      config: function(manager, req, res) {
        return manager.config.run(opts);
      },
      configTen: function(manager, req, res) {
        return manager.config.run(cloneOpts(opts, {
          tenant: req.params.id
        }));
      },
      reload: function(manager, req, res) {
        return manager.reload.run(opts);
      },
      upgrade: function(manager, req, res) {
        return manager.upgrade.run(opts);
      },
      check: function(manager, req, res) {
        return manager.check.run(opts);
      },
      log: function(manager, req, res) {
        return manager.log.run(argOpts(opts, ['log', req.body.msg]));
      },
      sping: function(manager, req, res) {
        return manager.sping.run(opts);
      },
      stop: function(manager, req, res) {
        return manager.stop.run(opts);
      },
      start: function(manager, req, res) {
        return manager.start.run(opts);
      }
    };
    commands = {};
    for (k in workers) {
      v = workers[k];
      loopfn = function(k, v) {
        return commands[k] = function(req, res) {
          var handler, manager;
          handler = responseHandler(req, res, output);
          manager = manage.commands(handler);
          return v(manager, req, res);
        };
      };
      loopfn(k, v);
    }
    return commands;
  };

  responseHandler = function(req, res, output) {
    var emitter, send;
    send = sendTo(res);
    emitter = new events.EventEmitter();
    emitter.on('success', function(obj) {
      return send(null, {
        success: _.assign({
          success: true
        }, obj)
      });
    });
    emitter.on('failure', function(obj) {
      return send({
        status: 500,
        error: obj
      });
    });
    emitter.on('part', function(...part) {
      return output.log(...part);
    });
    emitter.on('multi', function(multi) {
      var i, len, m, results;
      results = [];
      for (i = 0, len = multi.length; i < len; i++) {
        m = multi[i];
        results.push(output.log(...m));
      }
      return results;
    });
    return emitter;
  };

  sendTo = function(res) {
    return function(err, item) {
      var ref;
      if (err != null) {
        if (err.status != null) {
          res.status(err.status);
          err = err.error;
        } else {
          res.status(500);
        }
        return res.send({
          message: 'Data Error',
          error: err
        });
      } else {
        if (item == null) {
          return res.status(404).send({
            message: 'Not found'
          });
        } else if (item.failure != null) {
          return res.status((ref = item.status) != null ? ref : 500).send(item.failure);
        } else {
          return res.send(item.success);
        }
      }
    };
  };

  cloneOpts = function(opts, more) {
    var importOpts, opts2;
    opts2 = _.cloneDeep(opts);
    importOpts = function(obj, src) {
      var key, own;
      own = {}.hasOwnProperty;
      for (key in src) {
        if (own.call(src, key)) {
          obj[key] = src[key];
        }
      }
      return obj;
    };
    importOpts(opts2.options, more);
    return opts2;
  };

  argOpts = function(opts, argv) {
    var opts2;
    opts2 = _.cloneDeep(opts);
    opts2.argv = argv;
    return opts2;
  };

}).call(this);
