(function() {
  // tunnel.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // This is the main client interaction, which creates a reverse
  // (forward-in) ssh tunnel into the server for HTTP calls.
  // Similar to (e.g.) `ssh -N -R :remoteport:localhost:8080 user@g2c`
  var _, connect, emitter, events, iceptPath, net, open, ssh2, tls;

  _ = require('lodash');

  events = require('events');

  net = require('net');

  tls = require('tls');

  ssh2 = require('ssh2');

  emitter = require('../emitter');

  exports.iceptPath = iceptPath = '/-g2c-intercept';

  exports.create = function() {
    var protocol;
    protocol = new events.EventEmitter();
    protocol.connect = function(bt, tenant) {
      return connect(protocol, bt, tenant);
    };
    return protocol;
  };

  connect = function(protocol, bt, tenant) {
    protocol.config = bt.config;
    return open(protocol, tenant);
  };

  open = function(protocol, tenant) {
    var cfgData, conn, ended;
    cfgData = protocol.config;
    _.defaultsDeep(cfgData, {
      bindAddr: 'localhost',
      local: {
        host: 'localhost',
        port: 8080
      }
    });
    if (tenant == null) {
      tenant = {};
    }
    if (tenant.bindPort == null) {
      tenant.bindPort = 11080;
    }
    _.defaultsDeep(tenant, {
      bindPort: 11080,
      config: {
        local: {
          host: cfgData.local.host,
          port: cfgData.local.port,
          tls: cfgData.local.tls
        }
      }
    });
    ended = false;
    protocol.end = function() {};
    conn = new ssh2.Client();
    conn.on('error', function(err) {
      return protocol.emit('error', err);
    });
    conn.on('close', function(hadErr) {
      return protocol.emit('close', hadErr);
    });
    conn.on('ready', function() {
      return conn.forwardIn(cfgData.bindAddr, tenant.bindPort, function(err) {
        if (err != null) {
          protocol.emit('error', err);
          return conn.end();
        }
        protocol.emit('connected');
        return protocol.end = function() {
          if (!ended) {
            conn.end();
          }
          return ended = true;
        };
      });
    });
    conn.on('tcp connection', function(info, accept, reject) {
      var bindToLocal, channel, le;
      channel = null;
      bindToLocal = function(preamble) {
        var bound, local, ref, ref1;
        channel.unlisten();
        bound = false;
        if (tenant.config.local.tls) {
          local = tls.connect({
            port: tenant.config.local.port,
            host: tenant.config.local.host,
            rejectUnauthorized: false,
            servername: tenant.config.local.host,
            checkServerIdentity: function() {
              return void 0;
            }
          }, function() {
            if (preamble != null) {
              local.write(preamble);
            }
            local.pipe(channel).pipe(local);
            bound = true;
            return le.emit('bound');
          });
          local.on('error', function(err) {
            le.emit('error', err);
            if (channel != null) {
              channel.end();
            }
            return channel = null;
          });
          local.on('close', function(err) {
            if (bound) {
              le.emit('unbound');
            }
            if (channel != null) {
              channel.end();
            }
            return channel = null;
          });
          if (((ref = cfgData.qbmonitor) != null ? ref.local : void 0) != null) {
            return tenant.config.local = cfgData.qbmonitor.local;
          }
        } else {
          local = new net.Socket();
          local.on('error', function(err) {
            le.emit('error', err);
            if (channel != null) {
              channel.end();
            }
            return channel = null;
          });
          local.on('close', function(err) {
            if (bound) {
              le.emit('unbound');
            }
            if (channel != null) {
              channel.end();
            }
            return channel = null;
          });
          if (((ref1 = cfgData.qbmonitor) != null ? ref1.local : void 0) != null) {
            tenant.config.local = cfgData.qbmonitor.local;
          }
          return local.connect(tenant.config.local.port, tenant.config.local.host, function() {
            if (preamble != null) {
              local.write(preamble);
            }
            local.pipe(channel).pipe(local);
            bound = true;
            return le.emit('bound');
          });
        }
      };
      le = new events.EventEmitter();
      le.info = info;
      protocol.emit('received', le);
      if ((channel = accept()) != null) {
        emitter.augment(channel);
        channel.on('error', function(err) {
          return le.emit('error', err);
        });
        channel.listenOn('readable', function() {
          var best, buf, readCt, some, target;
          readCt = 0;
          best = `GET ${iceptPath}`;
          buf = Buffer.alloc(best.length);
          target = Buffer.from(best, 'utf8');
          if ((some = channel.read(best.length - readCt)) == null) {
            if ((some = channel.read()) != null) {
              some.copy(buf, readCt);
              readCt += some.length;
              if (!buf.equals(target.slice(0, readCt))) {
                return bindToLocal(buf.slice(0, readCt));
              }
            }
          } else {
            some.copy(buf, readCt);
            readCt += some.length;
            if (buf.length !== best.length) {
              return bindToLocal(buf.slice(0, readCt));
            } else {
              if (buf.equals(target)) {
                channel.unlisten();
                return channel.end(['HTTP/1.1 200 OK', 'Content-Length: 18', 'Connection: close', '', '{"success": true}\n'].join('\r\n'));
              } else {
                return bindToLocal(buf);
              }
            }
          }
        });
        return channel.listenOn('end', function() {
          return bindToLocal(null);
        });
      } else {
        return conn.end();
      }
    });
    return conn.connect({
      host: cfgData.host,
      port: cfgData.port || 22,
      username: cfgData.username,
      privateKey: cfgData.keydata,
      passphrase: cfgData.passphrase,
      keepaliveInterval: 30000,
      keepaliveCountMax: 3
    });
  };

}).call(this);
