(function() {
  // errors.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A utility to log and print out debug and error messages. This will
  // recognize common error types and output them in human readable form.
  var SEVERE, _, create, logger, mkdirp, path, pickLog, winston;

  _ = require('lodash');

  path = require('path');

  winston = require('winston');

  mkdirp = require('mkdirp');

  logger = winston.createLogger({
    transports: [new winston.transports.Console()]
  });

  exports.SEVERE = SEVERE = 10;

  exports.pickLog = pickLog = function(cfgpath, name, cb) {
    var logDirPath, logPath;
    logDirPath = path.join(path.dirname(cfgpath), 'logs');
    logPath = path.join(logDirPath, `${name}.log`);
    return mkdirp(logDirPath).then((err) => {
      return cb(err, logPath);
    });
  };

  exports.init = function(cfgpath, name, level, cb) {
    return pickLog(cfgpath, name, function(err, logPath) {
      var returnLevel;
      if (err != null) {
        return cb(err);
      }
      logger.add(new winston.transports.File({
        filename: logPath,
        maxsize: 1000000, // bytes
        maxFiles: 100,
        tailable: true
      }));
      returnLevel = 0;
      logger.level = (level = 'verbose') ? 'debug' : 'info';
      return cb(null, create(level));
    });
  };

  exports.create = create = function(level) {
    var debug, error, log, panic, warn;
    return {
      exit: function() {
        return process.exit(Math.max(0, returnLevel - SEVERE + 1));
      },
      level: level,
      debug: debug = logger.debug.bind(logger),
      log: log = logger.info.bind(logger),
      warn: warn = logger.warn.bind(logger),
      error: error = logger.error.bind(logger),
      panic: panic = function(...args) {
        error(...args);
        return process.exit(1);
      },
      setLevel: function(level) {
        return logger.level = level;
      },
      prettyConfig: function(level, err) {
        var returnLevel;
        returnLevel = Math.max(returnLevel, level);
        switch (err != null ? err.code : void 0) {
          case 'ENOENT':
            return warn(`Couldn't access file \"${err.path}\".`);
          default:
            return warn('Unexpected configuration error:', err);
        }
      },
      prettyListen: function(level, err, config) {
        var returnLevel;
        returnLevel = Math.max(returnLevel, level);
        switch (err != null ? err.code : void 0) {
          case 'EADDRINUSE':
            return warn(`Something already listening on port ${config.listenPort}.`);
          default:
            return warn('Unexpected command error', err);
        }
      },
      prettyBacktalk: function(level, err, config) {
        var ref, returnLevel;
        returnLevel = Math.max(returnLevel, level);
        switch (false) {
          case !_.isString(err):
            return warn(err);
          case (err != null ? err.code : void 0) !== 'ECONNREFUSED':
            return warn(`Unable to connect to server at \"${config.host}:${config.backPort}\".`);
          case (err != null ? (ref = err.message) != null ? ref.indexOf('CRYPT') : void 0 : void 0) === -1:
            return warn('Unexpected backchannel encryption error:', err);
          default:
            return warn('Unexpected backchannel error:', err);
        }
      },
      prettyTunnel: function(level, err, config) {
        var ref, returnLevel;
        returnLevel = Math.max(returnLevel, level);
        switch ((ref = err.error) != null ? ref.code : void 0) {
          case 'ECONNREFUSED':
            return warn(`Unable to connect to local \"${config.local.host}:${config.local.port}\".`);
          default:
            return warn("Unexpected tunnel error", err);
        }
      },
      prettyProxy: function(level, regId, err, config) {
        var returnLevel;
        returnLevel = Math.max(returnLevel, level);
        switch (err != null ? err.code : void 0) {
          case 'EADDRINUSE':
            return warn(`Unable to start: address ${config.proxyPort} already in use.`);
          default:
            if (regId != null) {
              return warn(`Unexpected proxy error for ${regId}`, err);
            } else {
              return warn("Unexpected proxy error", err);
            }
        }
      }
    };
  };

}).call(this);
