(function() {
  // listen.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A command listener for the client daemon, which allows the client
  // to receive and respond to simple instructions (start, stop, status)
  var _, events, listen, net, open;

  _ = require('lodash');

  events = require('events');

  net = require('net');

  exports.create = function() {
    var protocol;
    protocol = new events.EventEmitter();
    protocol.listen = function(config, defPort) {
      return listen(protocol, config, defPort);
    };
    protocol.end = function() {};
    return protocol;
  };

  listen = function(protocol, config, defPort) {
    protocol.status = {};
    protocol.config = config;
    return open(protocol, defPort);
  };

  open = function(protocol, defPort) {
    var ended, server;
    _.defaults(protocol.config, {
      listenPort: defPort || 3015
    });
    ended = false;
    server = new net.Server();
    server.on('error', function(err) {
      return protocol.emit('error', err);
    });
    server.on('close', function() {
      return protocol.emit('close');
    });
    server.on('listening', function() {
      protocol.end = function() {
        if (!ended) {
          server.close();
        }
        return ended = true;
      };
      return protocol.emit('listening', server.address());
    });
    server.on('connection', function(socket) {
      var draw, handle, readAccumulate, sp;
      sp = new events.EventEmitter();
      _.assign(sp, {
        reply: function(data) {
          return socket.write(JSON.stringify(data));
        },
        end: function(data) {
          if (data != null) {
            return socket.end(JSON.stringify(data));
          } else {
            return socket.end();
          }
        }
      });
      protocol.emit('connection', sp);
      socket.setEncoding('utf8');
      socket.on('error', function(err) {
        return sp.emit('error', err);
      });
      socket.on('close', function(hadErr) {
        return sp.emit('close', hadErr);
      });
      socket.on('data', function(data) {
        // TODO : data accumulator for long commands?
        sp.emit('data', data);
        return draw(data);
      });
      readAccumulate = '';
      draw = function(sdata) {
        var etx, rdata, results;
        readAccumulate += sdata;
        results = [];
        while ((etx = readAccumulate.indexOf('\x03')) !== -1) {
          rdata = readAccumulate.substr(0, etx);
          readAccumulate = readAccumulate.substr(etx + 1);
          results.push(handle(rdata));
        }
        return results;
      };
      return handle = function(sdata) {
        return protocol.emit('command', sdata.trim(), sp);
      };
    });
    return server.listen({
      port: protocol.config.listenPort
    });
  };

}).call(this);
