(function() {
  // lock.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A lock mechanism to ensure that a daemon only has a single instance.
  var bindExit, findDeep, fs, isRunning, lock, lockfile, path, unbindExit;

  lockfile = require('lockfile');

  fs = require('fs');

  path = require('path');

  isRunning = function(pid) {
    var e;
    try {
      process.kill(pid, 0);
    } catch (error) {
      e = error;
      return false;
    }
    return true;
  };

  exports.lock = lock = function(name, status, configPath, configData, cb) {
    var lockPath, pidPath, verification;
    lockPath = path.join(path.dirname(configPath), `${name}.lock`);
    pidPath = path.join(path.dirname(configPath), `${name}.pid`);
    verification = {lockPath, pidPath};
    return lockfile.lock(lockPath, {}, function(err) {
      var fail;
      if (err != null) {
        fail = function() {
          console.error(`File ${lockPath} already exists: is another instance running?`, err);
          return process.exit(1);
        };
        return fs.exists(pidPath, function(exists) {
          if (exists) {
            return fs.readFile(pidPath, {
              encoding: 'utf8'
            }, function(err, data) {
              var pidobj;
              if (err != null) {
                return cb(err);
              }
              pidobj = JSON.parse(data);
              if (!isRunning(pidobj.pid)) {
                fs.unlinkSync(pidPath);
                fs.unlinkSync(lockPath);
                return lock(name, status, configPath, configData, cb);
              } else {
                return fail();
              }
            });
          } else {
            return fail();
          }
        });
      } else {
        return fs.writeFile(pidPath, JSON.stringify(status), {
          encoding: 'utf8'
        }, function(err) {
          if (err) {
            console.error(`Could not write status file ${pidPath}`, err);
            return process.exit(1);
          } else {
            verification.group = bindExit(function() {
              fs.unlinkSync(pidPath);
              return lockfile.unlockSync(lockPath);
            });
            return cb(null, verification);
          }
        });
      }
    });
  };

  exports.unlock = function(verification) {
    fs.unlinkSync(verification.pidPath);
    lockfile.unlockSync(verification.lockPath);
    return unbindExit(verification.group);
  };

  exports.find = function(name, configPath, cb) {
    return findDeep(name, configPath, false, cb);
  };

  exports.findDeep = findDeep = function(name, configPath, erase, cb) {
    var lockPath, pidPath;
    lockPath = path.join(path.dirname(configPath), `${name}.lock`);
    pidPath = path.join(path.dirname(configPath), `${name}.pid`);
    return fs.exists(lockPath, function(exists) {
      if (!exists) {
        return cb(null, false);
      }
      return fs.exists(pidPath, function(exists) {
        if (!exists) {
          return cb(null, false);
        }
        return fs.readFile(pidPath, {
          encoding: 'utf8'
        }, function(err, data) {
          var pidobj;
          if (err != null) {
            return cb(err);
          }
          pidobj = JSON.parse(data);
          if (!erase) {
            return cb(null, true, pidobj);
          }
          if (isRunning(pidobj.pid)) {
            return cb(null, true, pidobj);
          } else {
            fs.unlinkSync(pidPath);
            fs.unlinkSync(lockPath);
            return cb(null, false);
          }
        });
      });
    });
  };

  bindExit = function(fn) {
    var group, handled;
    handled = false;
    group = {
      fn: fn,
      exit: function() {
        if (!handled) {
          fn();
        }
        return handled = true;
      },
      sigint: function() {
        if (!handled) {
          fn();
        }
        handled = true;
        return process.exit(130);
      },
      sigterm: function() {
        if (!handled) {
          fn();
        }
        handled = true;
        return process.exit(131);
      }
    };
    process.on('exit', group.exit);
    process.on('SIGINT', group.sigint);
    process.on('SIGTERM', group.sigterm);
    return group;
  };

  unbindExit = function(group) {
    process.removeListener('exit', group.exit);
    process.removeListener('SIGINT', group.sigint);
    process.removeListener('SIGTERM', group.sigterm);
    return group;
  };

}).call(this);
