(function() {
  // backtalk.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A backchannel listener for the server daemon, which allows the server
  // to receive and respond to backchannel handshakes.
  var _, async, config, createClientVitae, createServerVitae, crypto, events, forge, listen, net, open, path, ready, singleize;

  _ = require('lodash');

  async = require('async');

  crypto = require('crypto');

  events = require('events');

  forge = require('node-forge');

  net = require('net');

  path = require('path');

  config = require('../config');

  createServerVitae = function(cfgData, socket) {
    var ref, vitae;
    vitae = _.cloneDeep((ref = cfgData.vitae) != null ? ref : {});
    _.assign(vitae, {
      ip: socket.remoteAddress,
      version: config.version().version
    });
    return vitae;
  };

  createClientVitae = function(data, socket) {
    var ref, vitae;
    vitae = _.cloneDeep((ref = data.vitae) != null ? ref : {});
    _.assign(vitae, {
      ip: socket.remoteAddress,
      since: new Date().toString()
    });
    return vitae;
  };

  singleize = function(vals) {
    if (vals.length === 1) {
      return vals[0];
    } else {
      return vals;
    }
  };

  exports.create = function() {
    var protocol;
    protocol = new events.EventEmitter();
    protocol.listen = function(lt) {
      return listen(protocol, lt);
    };
    return protocol;
  };

  listen = function(protocol, lt) {
    protocol.status = {};
    protocol.config = lt.config;
    protocol.backchannels = {};
    return open(protocol);
  };

  open = function(protocol) {
    var ended, server;
    _.defaults(protocol.config, {
      backPort: 3015
    });
    ended = false;
    protocol.end = function() {};
    server = new net.Server();
    server.on('error', function(err) {
      return protocol.emit('error', err);
    });
    server.on('close', function() {
      return protocol.emit('close');
    });
    server.on('listening', function() {
      protocol.end = function() {
        if (!ended) {
          ended = true;
          server.close();
          return protocol.emit('shutdown');
        }
      };
      return protocol.emit('listening', server.address());
    });
    server.on('connection', function(socket) {
      var shutdownFn, sp, spended;
      shutdownFn = function() {
        return socket.end();
      };
      protocol.on('shutdown', shutdownFn);
      sp = new events.EventEmitter();
      spended = false;
      _.assign(sp, {
        reply: function(data) {
          return socket.write(JSON.stringify(data) + '\x03');
        },
        end: function() {
          if (!spended) {
            spended = true;
            return socket.end();
          }
        }
      });
      socket.setEncoding('utf8');
      socket.on('error', function(err) {
        return sp.emit('error', err);
      });
      socket.on('close', function(hadErr) {
        return protocol.removeListener('shutdown', shutdownFn);
      });
      socket.on('data', function(data) {
        return sp.emit('data', data);
      });
      ready(protocol, socket, sp);
      return protocol.emit('connection', sp);
    });
    return server.listen({
      port: protocol.config.backPort
    });
  };

  ready = function(protocol, socket, sp) {
    var bindPort, cfgData, closing, draw, handle, noted, patches, rawprompt, readAccumulate, registeredId, registry, responses;
    cfgData = protocol.config;
    registry = cfgData.registry;
    patches = cfgData.patches;
    rawprompt = null;
    registeredId = null;
    bindPort = null;
    closing = false;
    noted = [];
    readAccumulate = '';
    draw = function(sdata) {
      var etx, rdata, results;
      readAccumulate += sdata;
      results = [];
      while ((etx = readAccumulate.indexOf('\x03')) !== -1) {
        rdata = readAccumulate.substr(0, etx);
        readAccumulate = readAccumulate.substr(etx + 1);
        results.push(handle(rdata));
      }
      return results;
    };
    socket.on('data', draw);
    socket.on('close', function(hadErr) {
      var i, len, note, ref, ref1;
      closing = true;
      for (i = 0, len = noted.length; i < len; i++) {
        note = noted[i];
        if (!((ref = (ref1 = sp.stale) != null ? ref1[note] : void 0) != null ? ref : false)) {
          delete protocol.backchannels[note];
          registry.disconnected(note);
        }
      }
      sp.emit('close', hadErr, noted);
      return noted = [];
    });
    responses = {
      ping1: function(data) {
        return sp.reply({
          type: 'didPing1'
        });
      },
      check1: function(data) {
        var ref;
        return patches.check((ref = data.vitae) != null ? ref : {}, function(err, patch) {
          if (err != null) {
            return sp.reply({
              type: 'error',
              error: err
            });
          }
          return sp.reply({
            type: 'didCheck1',
            patch: patch
          });
        });
      },
      upgrade1: function(data) {
        return patches.retrieve(data.version, data.id, function(err, patch) {
          var giver;
          if (err != null) {
            return sp.reply({
              type: 'error',
              error: err
            });
          }
          giver = patches.readable(patch, function() {
            return socket.setEncoding('utf8');
          });
          // see https://github.com/nodejs/node/issues/6038
          socket.setEncoding(null);
          delete socket._writableState.encoder;
          return giver.pipe(socket);
        });
      },
      handshake1: function(data) {
        var replacement;
        replacement = [];
        registeredId = data.registeredId;
        if (!(_.isArray(registeredId))) {
          registeredId = [registeredId];
        }
        if (data.handshakeVersion !== 1) {
          return sp.reply({
            type: 'error',
            error: "Bad handshake version " //{data.handshakeVersion}\" != 1"
          });
        }
        return registry.sameKey(registeredId, function(err, sk) {
          if (err != null) {
            return sp.reply({
              type: 'error',
              error: `Couldn't detect keys in registry: ${err}`
            });
          }
          if (!sk) {
            return sp.reply({
              type: 'error',
              error: `Registered IDs ${registeredId} use different keys.`
            });
          }
          replacement = registry.isAvailable(registeredId);
          if (!(_.isEmpty(replacement))) {
            sp.emit('replacement', replacement);
          }
          registry.connecting(registeredId);
          return registry.find(registeredId, function(err, key, ports) {
            var cryptText;
            if (err != null) {
              return sp.reply({
                type: 'error',
                error: `No registration found for \"${registeredId}\"`
              });
            }
            bindPort = ports;
            rawprompt = crypto.randomBytes(50);
            cryptText = crypto.publicEncrypt(key, rawprompt);
            return sp.reply({
              type: 'shakehand1',
              replacement: replacement,
              prompt: cryptText.toString('base64')
            });
          });
        });
      },
      handshake2: function(data) {
        var vitae;
        if ((rawprompt == null) || !Buffer.from(data.response, 'base64').equals(rawprompt)) {
          return sp.reply({
            type: 'error',
            error: "Failed identity verification"
          });
        }
        // TODO: pick port and remember connection
        vitae = createServerVitae(cfgData, socket);
        return sp.reply({
          type: 'shakehand2',
          bindPort: singleize(bindPort),
          vitae: vitae
        });
      },
      handshake3: function(data) {
        var vitae;
        vitae = createClientVitae(data, socket);
        return async.each(registeredId, function(regId, cb) {
          var afterClose, closeDone, complete, oldsp;
          complete = function() {
            if (closing) {
              return;
            }
            registry.connected(regId);
            registry.setVitae(regId, vitae);
            protocol.backchannels[regId] = sp;
            noted.push(regId);
            return cb(null);
          };
          if ((oldsp = protocol.backchannels[regId]) != null) {
            closeDone = false;
            afterClose = function() {
              if (closeDone) {
                return;
              }
              closeDone = true;
              if (oldsp.stale == null) {
                oldsp.stale = {};
              }
              oldsp.stale[regId] = true;
              return complete();
            };
            oldsp.once('close', afterClose);
            setTimeout(afterClose, 1000);
            return oldsp.end();
          } else {
            return complete();
          }
        }, function(err) {
          return sp.emit('handshake', sp, registeredId);
        });
      },
      register1: function(data) {
        return registry.register(data.publicKey, data.preferred, function(err, regId, entry) {
          if (err != null) {
            return sp.reply({
              type: 'error',
              error: "Failed public key." + err.toString()
            });
          }
          return sp.reply({
            type: 'didRegister1',
            registeredId: regId,
            registeredPort: entry.port
          });
        });
      },
      register2: function(data) {
        return sp.emit('registered');
      },
      unregister1: function(data) {
        if (data.handshakeVersion !== 1) {
          return sp.reply({
            type: 'error',
            error: "Bad handshake version " //{data.handshakeVersion}\""
          });
        }
        registeredId = data.registeredId;
        if (!(_.isArray(registeredId))) {
          registeredId = [registeredId];
        }
        return registry.find(registeredId, function(err, key, ports) {
          var cryptText;
          if (err != null) {
            return sp.reply({
              type: 'error',
              error: `No registration found for \"${registeredId}\": ${err}`
            });
          }
          rawprompt = crypto.randomBytes(50);
          cryptText = crypto.publicEncrypt(key, rawprompt);
          return sp.reply({
            type: 'didUnregister1',
            prompt: cryptText.toString('base64')
          });
        });
      },
      unregister2: function(data) {
        if ((rawprompt == null) || !Buffer.from(data.response, 'base64').equals(rawprompt)) {
          return sp.reply({
            type: 'error',
            error: "Failed identity verification"
          });
        }
        return registry.unregister(registeredId, function(err) {
          if (err != null) {
            return sp.reply({
              type: 'error',
              error: `Unregistration failed for \"${registeredId}\": ${err}`
            });
          }
          return sp.reply({
            type: 'didUnregister2'
          });
        });
      },
      unregister3: function(data) {
        return sp.emit('unregistered');
      }
    };
    return handle = function(sdata) {
      var data, e, ref, respf;
      try {
        data = JSON.parse(sdata);
      } catch (error) {
        e = error;
        return protocol.emit('error', e);
      }
      respf = (ref = responses[data.type]) != null ? ref : function(data) {
        return sp.reply({
          type: 'error',
          error: `No response to type \"${data.type}\"`
        });
      };
      return respf(data);
    };
  };

}).call(this);
