#!/usr/bin/env node

(function() {
  //!/usr/bin/env coffee

  // reg-convert-bin.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A simple executable to move a registry from a file into redis.
  var _, async, fs, rclient, redis, regFn, regUrl, warn;

  _ = require('lodash');

  redis = require('redis');

  async = require('async');

  fs = require('fs');

  if (process.argv.length < 3) {
    return console.warn("Usage: reg-convert <registryName>");
  }

  regFn = process.argv[2];

  regUrl = 'redis://localhost:6379';

  rclient = redis.createClient(regUrl);

  warn = function(...args) {
    console.warn(...args);
    return rclient.quit();
  };

  fs.readFile(regFn, {
    encoding: 'utf8'
  }, function(err, data) {
    var ex, max, maxPort, ref, ref1, registry;
    if (err != null) {
      return warn(`Unable to read ${regFn}: ${err}`);
    }
    try {
      registry = JSON.parse(data);
    } catch (error) {
      ex = error;
      if (err != null) {
        return warn(`Unable to parse ${regFn}: ${err}`);
      }
    }
    max = Math.max(4000, (ref = registry.maxRegisteredId) != null ? ref : 0);
    maxPort = Math.max(14000, (ref1 = registry.maxRegisteredPort) != null ? ref1 : 0);
    return async.eachSeries(Object.keys(registry.registeredIds), function(regId, cb) {
      var key, port, vitae;
      ({port, key, vitae} = registry.registeredIds[regId]);
      if (/^\d+$/.test(regId)) {
        max = Math.max(max, parseInt(regId));
      }
      return async.series([
        function(cb) {
          if (vitae != null) {
            return rclient.set(`g2c/ids/${regId}/vitae`,
        JSON.stringify(vitae),
        cb);
          } else {
            return cb();
          }
        },
        function(cb) {
          return rclient.set(`g2c/ids/${regId}/key`,
        key,
        cb);
        },
        function(cb) {
          return rclient.set(`g2c/ids/${regId}/port`,
        port,
        cb);
        },
        function(cb) {
          return rclient.set(`g2c/ports/${port}`,
        regId,
        cb);
        },
        function(cb) {
          return rclient.sadd("g2c/allids",
        regId,
        cb);
        }
      ], cb);
    }, function(err) {
      return async.series([
        function(cb) {
          return rclient.set("g2c/max/port",
        `${maxPort}`,
        cb);
        },
        function(cb) {
          return rclient.set("g2c/max/id",
        `${max}`,
        cb);
        }
      ], function(err) {
        console.log("Done");
        return rclient.quit();
      });
    });
  });

}).call(this);
