#!/usr/bin/env node

(function() {
  //!/usr/bin/env coffee

  // petit-server-bin.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.
  var _, backtalk, childs, commands, config, forge, fs, getopt, lock, main, net, patches, path, registry, say, util, warn;

  _ = require('lodash');

  childs = require('child_process');

  getopt = require('node-getopt');

  forge = require('node-forge');

  fs = require('fs');

  net = require('net');

  path = require('path');

  config = require('../config');

  lock = require('../lock');

  backtalk = require('./backtalk');

  registry = require('./registry_redis');

  patches = require('./patches');

  util = require('./util');

  say = console.log;

  warn = console.warn;

  main = function() {
    var opt, opts, optsrc, ref, ref1, task, usage;
    usage = function() {
      say("Usage: petit-server <command> [OPTIONS]");
      say("Known commands are:");
      say("  " + _.sortBy(_.keys(commands)).join(', '));
      return process.exit(1);
    };
    if (process.argv.length < 3 || ((ref = process.argv[2]) === '-h' || ref === '--help')) {
      usage();
    }
    task = commands[process.argv[2]];
    if (task == null) {
      say(`Unknown command \"${process.argv[2]}\"`);
      usage();
    }
    optsrc = (ref1 = task.opts) != null ? ref1 : [];
    optsrc.push(...[['c', 'config=ARG', 'use config file'], ['v', 'verbose', 'print debug output'], ['h', 'help', 'display this help']]);
    opt = getopt.create(optsrc);
    opt.setHelp(`Usage: petit-server ${process.argv[2]} [OPTION]

[[OPTIONS]]`);
    opts = opt.bindHelp().parseSystem();
    return task.run(opts);
  };

  commands = {
    init: {
      opts: [],
      run: function(opts) {
        return config.createConfig(opts.options.config, function(err, made, path) {
          if (err != null) {
            return warn("Can't create config file.", err);
          }
          if (made) {
            say(`Created configuration file \"${path}\".`);
            return config.readConfig(path, function(err, configPath, configData) {
              if (err != null) {
                return warn(`Can't read config \"${configPath}\".`, err);
              }
              return registry.init(configPath, configData, function(err, made, path) {
                if (err != null) {
                  return warn("Can't create registry.", err);
                }
                say(`Created registry file \"${path}\".`);
                return patches.init(configPath, configData, function(err, made, path) {
                  if (err != null) {
                    return warn("Can't create patches.", err);
                  }
                  return say(`Created patches file \"${path}\".`);
                });
              });
            });
          } else {
            return say(`Configuration file \"${path}\" already exists.`);
          }
        });
      }
    },
    start: {
      opts: [],
      run: function(opts) {
        var configPath;
        configPath = config.pickConfig(opts.options.config);
        return fs.exists(configPath, function(exists) {
          if (!exists) {
            return warn(`No config at \"${configPath}\": use \"init\".`);
          }
          return config.readConfig(configPath, function(err, configPath, data) {
            var onExists, regPath, regUrl;
            if (err != null) {
              return warn(`Couldn't read config \"${configPath}\".`, err);
            }
            regPath = data.server.registryfile;
            regUrl = data.server.registryUrl;
            if ((regPath == null) && (regUrl == null)) {
              return warn("No registry file in config");
            }
            onExists = function(exists) {
              var exe;
              if (!exists) {
                say("No registry. Use \"init\" first.");
                process.exit(101);
              }
              exe = path.join(path.dirname(fs.realpathSync(__filename)), 'petit-serverd-bin') + path.extname(__filename);
              return lock.findDeep('server', configPath, true, function(err, exists, data) {
                var ch, checkgo;
                if (err != null) {
                  return warn(err);
                }
                if (exists) {
                  say(`Server is already running with pid ${data.pid}.`);
                  return process.exit(101);
                }
                ch = childs.spawn(exe, ['-c', configPath], {
                  detached: true,
                  stdio: 'ignore'
                });
                ch.once('error', function(err) {
                  return warn("Unable to start server.", err);
                });
                ch.unref();
                checkgo = function(bads) {
                  return function() {
                    return lock.find('server', configPath, function(err, exists, data) {
                      if (err != null) {
                        return warn(err);
                      }
                      if (exists) {
                        return say(`Server started with pid ${data.pid}.`);
                      } else if (bads < 20) {
                        return setTimeout(checkgo(bads + 1), 300);
                      } else {
                        return say("Unable to start server.");
                      }
                    });
                  };
                };
                // TODO : more often / dynamic polling
                return setTimeout(checkgo(0), 300);
              });
            };
            if (regUrl == null) {
              return fs.exists(regPath, onExists);
            } else {
              return onExists(true);
            }
          });
        });
      }
    },
    stop: {
      opts: [],
      run: function(opts) {
        var configPath, isKilled, waitKilled;
        configPath = config.pickConfig(opts.options.config);
        lock.find('server', configPath, function(err, exists, data) {
          var conn;
          if (err != null) {
            return warn(err);
          }
          if (!exists) {
            say("Server is not running.");
            return process.exit(101);
          }
          conn = new net.Socket();
          conn.setEncoding('utf8');
          conn.on('close', function() {
            return waitKilled(data.pid, function(killed) {
              if (killed) {
                return say("Server is stopped.");
              }
              process.kill(data.pid, 'SIGINT');
              return waitKilled(data.pid, function(killed) {
                if (killed) {
                  return say("Server is stopped with SIGINT.");
                } else {
                  return say(`Unable to stop server pid ${data.pid}`);
                }
              });
            });
          });
          return conn.connect(data.listen, 'localhost', function() {
            return conn.end('STOP\x03');
          });
        });
        waitKilled = function(pid, cb) {
          var maybeDone, repeater, total;
          if (isKilled(pid)) {
            return cb(true);
          }
          total = 0;
          maybeDone = function() {
            if (isKilled(pid)) {
              clearInterval(repeater);
              return cb(true);
            } else if (total < 10) {
              return ++total;
            } else {
              clearInterval(repeater);
              return cb(false);
            }
          };
          return repeater = setInterval(maybeDone, 100);
        };
        return isKilled = function(pid) {
          var e;
          try {
            process.kill(pid, 0);
          } catch (error) {
            e = error;
            return true;
          }
          return false;
        };
      }
    },
    status: {
      opts: [],
      run: function(opts) {
        var configPath;
        configPath = config.pickConfig(opts.options.config);
        return lock.find('server', configPath, function(err, exists, piddata) {
          var buf, conn;
          if (err != null) {
            return warn(err);
          }
          if (!exists) {
            say("Server is not running.");
            return process.exit(101);
          }
          conn = new net.Socket();
          conn.setEncoding('utf8');
          buf = "";
          conn.on('data', function(data) {
            var alias, cconn, e, fp, i, j, len, len1, obj, others, out, ref, ref1, ref10, ref11, ref12, ref13, ref14, ref15, ref2, ref3, ref4, ref5, ref6, ref7, ref8, ref9, regId;
            try {
              buf += data;
              obj = JSON.parse(buf);
            } catch (error) {
              e = error;
              // wait for more
              say("...retrieving data...");
              return;
            }
            say(`Server has status \"${obj.status}\" on pid ${piddata.pid}.`);
            ref = obj.connections;
            for (i = 0, len = ref.length; i < len; i++) {
              cconn = ref[i];
              regId = cconn.registeredId;
              if (opts.argv.length > 1 && regId !== opts.argv[1]) {
                continue;
              }
              fp = (ref1 = cconn.fingerprint) != null ? ref1 : '';
              if (cconn.connected === true) {
                say(` + ${regId} ${(ref2 = cconn.port) != null ? ref2 : '     '} ${fp}`);
              } else if (cconn.connected === false) {
                say(` - ${regId} ${(ref3 = cconn.port) != null ? ref3 : '     '} ${fp}`);
              } else {
                say(` / ${regId} ${(ref4 = cconn.port) != null ? ref4 : '     '} ${fp}`);
              }
              if (opts.argv.length > 1) {
                if (((ref5 = cconn.vitae) != null ? ref5.ip : void 0) != null) {
                  if (((ref6 = cconn.vitae) != null ? ref6.hostname : void 0) != null) {
                    say(`      \"${cconn.vitae.hostname}\" @ ${cconn.vitae.ip}`);
                  } else {
                    say(`      @ ${cconn.vitae.ip}`);
                  }
                } else if (((ref7 = cconn.vitae) != null ? ref7.hostname : void 0) != null) {
                  say(`  \"${cconn.vitae.hostname}\"`);
                }
                if (((ref8 = cconn.vitae) != null ? ref8.since : void 0) != null) {
                  say(`      ${cconn.vitae.since}`);
                }
                if (((ref9 = cconn.vitae) != null ? ref9.platform : void 0) != null) {
                  say(`      ${(ref10 = cconn.vitae.version) != null ? ref10 : 'unknown_version'} (${(ref11 = cconn.vitae.brand) != null ? ref11 : 'no_brand'}) on ${cconn.vitae.platform} ${(ref12 = cconn.vitae.arch) != null ? ref12 : 'unknown_arch'} ${(ref13 = cconn.vitae.release) != null ? ref13 : 'unknown_release'}`);
                }
                others = _.omit(cconn.vitae, ['ip', 'hostname', 'since', 'platform', 'arch', 'brand', 'release', 'version']);
                if (!(_.isEmpty(others))) {
                  out = _.map(_.toPairs(others), function([k, v]) {
                    return `${k}=${v}`;
                  }).join(", ");
                  say(`      ${out}`);
                }
                ref15 = (ref14 = cconn.aliases) != null ? ref14 : [];
                for (j = 0, len1 = ref15.length; j < len1; j++) {
                  alias = ref15[j];
                  say(`      = ${alias}`);
                }
              }
            }
            return conn.end();
          });
          if (opts.argv.length > 1) {
            return conn.connect(piddata.listen, 'localhost', function() {
              return conn.write(`STATUS ${opts.argv[1]}\x03`);
            });
          } else {
            return conn.connect(piddata.listen, 'localhost', function() {
              return conn.write('STATUS\x03');
            });
          }
        });
      }
    },
    addalias: {
      opts: [],
      run: function(opts) {
        var alias, configPath, regId;
        if (opts.argv.length !== 3) {
          say("Usage: petit-server addalias [OPTIONS] id alias");
          return process.exit(1);
        }
        regId = opts.argv[1].trim();
        alias = opts.argv[2].trim();
        configPath = config.pickConfig(opts.options.config);
        return lock.find('server', configPath, function(err, exists, piddata) {
          var conn;
          if (err != null) {
            return warn(err);
          }
          if (!exists) {
            say("Server is not running.");
            return process.exit(101);
          }
          conn = new net.Socket();
          conn.setEncoding('utf8');
          conn.on('data', function(data) {
            var obj;
            obj = JSON.parse(data);
            if (obj.success) {
              say("Alias added.");
            } else {
              say("Alias not added:", obj.error);
            }
            return conn.end();
          });
          return conn.connect(piddata.listen, 'localhost', function() {
            return conn.write(`ADDALIAS ${regId} ${alias}\x03`);
          });
        });
      }
    },
    removealias: {
      opts: [],
      run: function(opts) {
        var alias, configPath, regId;
        if (opts.argv.length !== 3) {
          say("Usage: petit-server addalias [OPTIONS] id alias");
          return process.exit(1);
        }
        regId = opts.argv[1].trim();
        alias = opts.argv[2].trim();
        configPath = config.pickConfig(opts.options.config);
        return lock.find('server', configPath, function(err, exists, piddata) {
          var conn;
          if (err != null) {
            return warn(err);
          }
          if (!exists) {
            say("Server is not running.");
            return process.exit(101);
          }
          conn = new net.Socket();
          conn.setEncoding('utf8');
          conn.on('data', function(data) {
            var obj;
            obj = JSON.parse(data);
            if (obj.success) {
              say("Alias deleted.");
            } else {
              say("Alias not deleted:", obj.error);
            }
            return conn.end();
          });
          return conn.connect(piddata.listen, 'localhost', function() {
            return conn.write(`DELALIAS ${regId} ${alias}\x03`);
          });
        });
      }
    },
    config: {
      opts: [],
      run: function(opts) {
        return config.readConfig(opts.options.config, function(err, cpath, data) {
          var onExists, regPath, regUrl;
          if (err != null) {
            return warn(err);
          }
          regPath = data.server.registryfile;
          regUrl = data.server.registryUrl;
          if ((regPath == null) && (regUrl == null)) {
            return warn("No registry in config.");
          }
          onExists = function(exists) {
            if (exists) {
              return util.verifyConfig(cpath, data, function(err) {
                if (err) {
                  return warn(err);
                }
                registry = data.server.registry;
                delete data.server.registry;
                return registry.size(function(err, sz) {
                  if (err != null) {
                    return warn(err);
                  }
                  if (sz > 0) {
                    say("Registry (configured):");
                    registry.each((function(err, {port, id, fingerprint, alss}) {
                      var alias, i, len, results;
                      if (err != null) {
                        return warn("Registry err: ", err);
                      } else {
                        say(`  id: ${id}, fingerprint: ${fingerprint}`);
                        if (alss != null ? alss.length : void 0) {
                          results = [];
                          for (i = 0, len = alss.length; i < len; i++) {
                            alias = alss[i];
                            results.push(say(`      a.k.a. ${alias}`));
                          }
                          return results;
                        }
                      }
                    }), function() {
                      return registry.quit();
                    });
                  } else {
                    registry.quit();
                    say("Server registry is empty.");
                  }
                  say(`Verified configuration file at ${cpath}.`);
                  return say("\"config\":", JSON.stringify(data.server, null, 2));
                });
              });
            } else {
              say("This server has no registry.");
              say(`Verified configuration file at ${cpath}.`);
              return say("\"config\":", JSON.stringify(data.server, null, 2));
            }
          };
          if (regUrl == null) {
            return fs.exists(regPath, onExists);
          } else {
            return onExists(true);
          }
        });
      }
    },
    version: {
      opts: [],
      run: function(opts) {
        var obj;
        obj = config.version();
        if ((obj.display != null) && obj.version !== obj.display) {
          say(`Ground2Cloud server version ${obj.version} (d.b.a. ${obj.display})`);
        } else {
          say(`Ground2Cloud server version ${obj.version}`);
        }
        return say("by Cloud Elements <cloud-elements.com>");
      }
    },
    verifypatches: {
      opts: [],
      run: function(opts) {
        return config.readConfig(opts.options.config, function(err, path, data) {
          var onExists, regPath, regUrl;
          if (err != null) {
            return warn(err);
          }
          regPath = data.server.registryfile;
          regUrl = data.server.registryUrl;
          if ((regPath == null) && (regUrl == null)) {
            return warn("No registry in config.");
          }
          onExists = function(exists) {
            if (exists) {
              return util.verifyConfig(path, data, function(err) {
                registry = data.server.registry;
                registry.quit();
                if (err) {
                  return warn(err);
                }
                patches = data.server.patches;
                return patches.verifyAll(function(err) {
                  if (err != null) {
                    return warn("Bad patches", err);
                  }
                  return say("Good patches");
                });
              });
            }
          };
          if (regUrl == null) {
            return fs.exists(regPath, onExists);
          } else {
            return onExists(true);
          }
        });
      }
    }
  };

  main();

}).call(this);
