#!/usr/bin/env node

(function() {
  //!/usr/bin/env coffee

  // petit-serverd-bin.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // This is the executable for the server-side of petit

  // petit-serverd [options]
  var SEVERE, _, cfgPath, config, errors, getopt, handle, handleConfig, level, lock, main, opt, output, restapi, server, startRest, util;

  _ = require('lodash');

  getopt = require('node-getopt');

  config = require('../config');

  errors = require('../errors');

  lock = require('../lock');

  server = require('./server');

  util = require('./util');

  restapi = require('./restapi');

  opt = getopt.create([['v', 'verbose', 'print debug output'], ['c', 'config=ARG', 'set config file'], ['h', 'help', 'display this help']]).bindHelp().parseSystem();

  SEVERE = errors.SEVERE;

  cfgPath = config.pickConfig(opt.options.config);

  level = opt.options.verbose ? 'verbose' : 'normal';

  output = null;

  main = function() {
    return config.readConfig(opt.options.config, handleConfig);
  };

  handleConfig = function(err, configPath, configData) {
    if (err != null) {
      return output.panic(`Error reading config \"${configPath}\":`, err);
    }
    return util.verifyConfig(configPath, configData, function(err) {
      if (err != null) {
        return output.prettyConfig(SEVERE, err);
      }
      return lock.lock('server', {
        pid: process.pid,
        listen: configData.server.listenPort
      }, configPath, configData, function(err, v) {
        var ref, sv;
        sv = server.create();
        if ((ref = configData.server) != null) {
          ref.configPath = configData.configPath;
        }
        handle(sv, configData.server);
        return sv.start(configData.server);
      });
    });
  };

  handle = function(sv, cdata) {
    sv.on('error', function() {}); // handle in sub-structures
    sv.on('start', function() {
      return output.log('Server started');
    });
    sv.on('close', function(hadErr) {
      return output.log('Server closed');
    });
    sv.on('command', function(cmd, c) {
      var alias, regId;
      output.log('Server command', cmd);
      switch (false) {
        case !_.startsWith(cmd, 'STATUS'):
          [cmd, regId] = cmd.split(' ');
          if (regId != null) {
            return sv.statusOf(regId, function(err, stt) {
              if (err) {
                return c.end({
                  success: false,
                  error: err.toString()
                });
              } else {
                return c.end(stt);
              }
            });
          } else {
            return sv.status(function(err, stt) {
              if (err) {
                return c.end({
                  success: false,
                  error: err.toString()
                });
              } else {
                return c.end(stt);
              }
            });
          }
          break;
        case cmd !== 'STOP':
          return sv.end();
        case !_.startsWith(cmd, 'ADDALIAS'):
          [cmd, regId, alias] = cmd.split(' ');
          return cdata.registry.addAlias(regId, alias, function(err) {
            if (err != null) {
              return c.end({
                success: false,
                error: err.toString()
              });
            } else {
              return c.end({
                success: true
              });
            }
          });
        case !_.startsWith(cmd, 'DELALIAS'):
          [cmd, regId, alias] = cmd.split(' ');
          return cdata.registry.removeAlias(regId, alias, function(err) {
            if (err != null) {
              return c.end({
                success: false,
                error: err.toString()
              });
            } else {
              return c.end({
                success: true
              });
            }
          });
        default:
          return output.warn(`Unknown command ${cmd}`);
      }
    });
    sv.listener.on('error', function(err) {
      return output.prettyListen(SEVERE, err, cdata);
    });
    sv.listener.on('close', function() {
      return output.log('Command listener closed.');
    });
    sv.listener.on('connection', function(c) {
      output.log('Command stream opened.');
      c.on('error', function(err) {
        return output.prettyListen(SEVERE, err, cdata);
      });
      return c.on('close', function() {
        return output.log('Command stream closed.');
      });
    });
    // c.on 'data', (data) -> output.log 'Command stream data', data
    sv.listener.on('listening', function(addr) {
      return output.log('Command listening on', addr);
    });
    sv.backchannel.on('error', function(err) {
      return output.prettyBacktalk(SEVERE, err, cdata);
    });
    sv.backchannel.on('close', function() {
      return output.log('Backchannel closed.');
    });
    sv.backchannel.on('listening', function(addr) {
      return output.log('Backchannel listening on', addr);
    });
    sv.backchannel.on('connection', function(c) {
      output.log('Backchannel stream opened.');
      c.on('error', function(err) {
        return output.prettyBacktalk(SEVERE, err, cdata);
      });
      c.on('close', function(hadErr, rid) {
        if (rid != null) {
          return output.log(`Backchannel ${rid} stream closed.`);
        } else {
          return output.log("Backchannel stream closed.");
        }
      });
      c.on('handshake', function(sp, rid) {
        return output.log(`Backchannel ${rid} handshake complete.`);
      });
      return c.on('replacement', function(rid) {
        return output.log(`Replacing regID ${rid}.`);
      });
    });
    // c.on 'data', (data) -> output.log 'Backchannel stream data:', data
    sv.backchannel.on('registered', function() {
      return output.log('Registered.');
    });
    sv.backchannel.on('unregistered', function() {
      return output.log('Unregistered.');
    });
    sv.proxy.on('nokey', function() {
      return output.log('No https proxy: no key file.');
    });
    sv.proxy.on('error', function(err) {
      return output.prettyProxy(SEVERE, null, err, cdata);
    });
    sv.proxy.on('close', function(err) {
      return output.log('Proxy closed.');
    });
    sv.proxy.on('listening', function() {
      return output.log('Proxy listening.');
    });
    return sv.proxy.on('connection', function(c, regId) {
      output.log(`Proxy stream for ${regId} opened.`);
      c.on('error', function(err) {
        return output.prettyProxy(SEVERE, regId, err, cdata);
      });
      return c.on('close', function(hadErr) {
        return output.log(`Proxy stream for ${regId} closed.`);
      });
    });
  };

  startRest = function() {
    return config.readConfig(opt.options.config, function(err, cfgPath, cfgData) {
      var apiPort, ref, ref1;
      if (err != null) {
        return output.warn(`Can't start REST API: ${err}`);
      }
      apiPort = (ref = (ref1 = cfgData.server) != null ? ref1.apiPort : void 0) != null ? ref : 8101;
      output.log(`Starting REST API server on ${apiPort}`);
      return restapi.start(apiPort, opt, output);
    });
  };

  errors.init(cfgPath, 'server', level, function(err, out) {
    if (err != null) {
      return console.error('Unable to initialize logger', err);
    }
    output = out;
    startRest();
    return main();
  });

}).call(this);
