(function() {
  // proxy.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A proxy server, which is able to forward http requests to their
  // correct localhost port (which will be remote listening on behalf of a
  // client)
  var _, async, createTlsOpts, crypto, events, extractHost, findEntry, fs, http, https, listen, open, path, proxy;

  async = require('async');

  path = require('path');

  fs = require('fs');

  _ = require('lodash');

  events = require('events');

  http = require('http');

  https = require('https');

  proxy = require('http-proxy');

  crypto = require('crypto');

  exports.create = function() {
    var protocol;
    protocol = new events.EventEmitter();
    protocol.listen = function(lt, bt) {
      return listen(protocol, lt, bt);
    };
    return protocol;
  };

  listen = function(protocol, lt, bt) {
    protocol.config = bt.config;
    return open(lt, bt, protocol);
  };

  open = function(lt, bt, protocol) {
    var createListen;
    _.defaults(protocol.config, {
      proxyPort: 8090,
      secureProxyPort: 8453
    });
    createListen = function(isSec, isLocalSec, tlsOpts) {
      var handler, px, registry, target, web;
      target = !isLocalSec ? function(port) {
        return `http://127.0.0.1:${port}`;
      } : function(port) {
        return `https://127.0.0.1:${port}`;
      };
      registry = protocol.config.registry;
      handler = function(req, res) {
        var cpt, host, port;
        cpt = new events.EventEmitter();
        req.on('error', function(err) {
          return cpt.emit('error', err);
        });
        res.on('error', function(err) {
          return cpt.emit('error', err);
        });
        req.on('end', function() {
          return cpt.emit('close', false);
        });
        req.on('close', function() {
          return cpt.emit('close', true);
        });
        [host, port] = extractHost(req.headers.host, 80);
        if (host == null) {
          protocol.emit('error', "No host header in proxy request");
          res.write(JSON.stringify({
            success: false,
            msg: "Proxy error: No host header"
          }));
          return res.end();
        }
        return findEntry(protocol.config, host, function(err, regId, entry) {
          var brand, ref, ref1, ref2;
          brand = (ref = entry != null ? (ref1 = entry.vitae) != null ? ref1.brand : void 0 : void 0) != null ? ref : 'ce';
          req.extras = {
            regId: regId,
            port: (ref2 = entry != null ? entry.port : void 0) != null ? ref2 : 0,
            brand: brand
          };
          if (err != null) {
            protocol.emit('error', err);
            res.statusCode = 404;
            console.log('recording unavailable for', regId);
            registry.used(regId, brand, new Date().getTime(), {
              success: false,
              reason: "service unavailable",
              statusCode: res.statusCode
            });
            res.write(JSON.stringify({
              g2c: 'unavailable'
            }));
            return res.end();
          }
          protocol.emit('connection', cpt, regId);
          return px.web(req, res, {
            target: target(entry.port),
            secure: false
          });
        });
      };
      px = proxy.createProxyServer();
      web = !isSec ? http.createServer(handler) : https.createServer(tlsOpts, handler);
      px.on('error', function(err, req, res) {
        var brand, ref, regId;
        protocol.emit('error', err);
        ({regId, brand} = (ref = req.extras) != null ? ref : {});
        res.statusCode = 500;
        registry.used(regId, brand, new Date().getTime(), {
          success: false,
          reason: "service unavailable",
          statusCode: res.statusCode
        });
        res.write(JSON.stringify({
          g2c: 'unavailable'
        }));
        return res.end();
      });
      px.on('proxyRes', function(proxyRes, req, res) {
        var brand, ref, regId;
        if ((regId = (ref = req.extras) != null ? ref.regId : void 0) != null) {
          brand = req.extras.brand;
          if (proxyRes.statusCode === 200) {
            return registry.used(regId, brand, new Date().getTime(), {
              success: true,
              req: {
                headers: req.headers
              },
              res: {
                headers: proxyRes.headers
              },
              statusCode: proxyRes.statusCode
            });
          } else {
            return registry.used(regId, brand, new Date().getTime(), {
              success: false,
              reason: "service error",
              req: {
                headers: req.headers
              },
              res: {
                headers: proxyRes.headers
              },
              statusCode: proxyRes.statusCode
            });
          }
        } else {
          return protocol.emit('warn', "Couldn't find request regId to record");
        }
      });
      web.on('error', function(err) {
        return protocol.emit('error', err);
      });
      web.on('listening', function() {
        protocol.on('ending', function() {
          return web.close();
        });
        return protocol.emit('listening');
      });
      return web.listen(!isSec ? protocol.config.proxyPort : protocol.config.secureProxyPort);
    };
    return createTlsOpts(protocol, protocol.config, function(err, tlsOpts) {
      if (err != null) {
        return protocol.emit('error', err);
      }
      protocol.end = function() {
        return protocol.emit('ending');
      };
      createListen(false, false);
      if (tlsOpts != null) {
        return createListen(true, false, tlsOpts);
      } else {
        return protocol.emit('nokey');
      }
    });
  };

  extractHost = function(hostport, defport) {
    var colon;
    if (hostport == null) {
      return [null, defport];
    }
    colon = hostport.indexOf(':');
    if (colon === -1) {
      return [hostport, defport];
    } else {
      return [hostport.substr(0, colon), parseInt(hostport.substr(colon + 1))];
    }
  };

  findEntry = function(config, host, cb) {
    var registry, testre;
    testre = /^\d+$/;
    registry = config.registry;
    return registry.findAlias(host, function(err, regId) {
      var dash, dot;
      if (err != null) {
        return cb(err);
      }
      if (!regId) {
        dash = host.indexOf('-');
        if (dash > -1) {
          regId = host.substring(0, dash);
          if (!testre.test(regId)) {
            regId = null;
          }
        }
        if (!regId) {
          dot = host.indexOf('.');
          if (dot > -1) {
            regId = host.substring(0, dot);
            if (!testre.test(regId)) {
              return cb(new Error(`Couldn't determine registeredId from ${host}`));
            }
          } else {
            return cb(new Error(`Couldn't determine registeredId from ${host}`));
          }
        }
      }
      return registry.get(regId, function(err, entry) {
        if (err != null) {
          return cb(err, regId);
        }
        if (registry.isConnected(regId)) {
          return cb(null, regId, entry);
        } else {
          return cb(new Error(`${regId} is not running.`), regId);
        }
      });
    });
  };

  createTlsOpts = function(protocol, config, cb) {
    var certDir, ciphers, honorCipherOrder, keyPath, secureOptions;
    // mozilla-suggested cipher block (allows some non-PFS cipers)
    ciphers = 'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-AES256-GCM-SHA384:DHE-RSA-AES128-GCM-SHA256:DHE-DSS-AES128-GCM-SHA256:kEDH+AESGCM:ECDHE-RSA-AES128-SHA256:ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA:ECDHE-ECDSA-AES128-SHA:ECDHE-RSA-AES256-SHA384:ECDHE-ECDSA-AES256-SHA384:ECDHE-RSA-AES256-SHA:ECDHE-ECDSA-AES256-SHA:DHE-RSA-AES128-SHA256:DHE-RSA-AES128-SHA:DHE-DSS-AES128-SHA256:DHE-RSA-AES256-SHA256:DHE-DSS-AES256-SHA:DHE-RSA-AES256-SHA:AES128-GCM-SHA256:AES256-GCM-SHA384:HIGH:!aNULL:!eNULL:!EXPORT:!DES:!3DES:!MD5:!PSK';
    secureOptions = crypto.constants.SSL_OP_NO_TLSv1 | crypto.constants.SSL_OP_NO_TLSv1_1;
    honorCipherOrder = true;
    certDir = path.join(path.dirname(config.configPath), 'certificates');
    keyPath = path.join(certDir, 'server.key');
    return fs.exists(keyPath, function(exists) {
      if (!exists) {
        return cb(null, null);
      }
      return fs.readFile(keyPath, function(err, key) {
        if (err != null) {
          return cb(`No key file certificates/server.key: ${err}`);
        }
        return fs.readFile(path.join(certDir, 'server.crt'), function(err, cert) {
          if (err != null) {
            return cb(`No certificate certificates/server.crt: ${err}`);
          }
          return fs.readFile(path.join(certDir, 'ca.bundle'), function(err, ca) {
            return cb(err, {secureOptions, key, cert, ciphers, honorCipherOrder, ca});
          });
        });
      });
    });
  };

  // async.map [capem1, capem2],
//  (capath, cb) -> fs.readFile path.join(certDir, capath), cb
//  (err, ca) -> cb err, {key, cert, ciphers, honorCipherOrder: true, ca}

}).call(this);
