(function() {
  // registry.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A registration utility to provide registry management functions for
  // the server.
  var _, akPath, async, createAuthkeys, createManager, forge, fs, home, mkdirp, path, shPath;

  _ = require('lodash');

  async = require('async');

  fs = require('fs');

  path = require('path');

  mkdirp = require('mkdirp');

  forge = require('node-forge');

  home = process.env.HOME || process.env.USERPROFILE;

  shPath = path.join(home, '.ssh');

  akPath = path.join(shPath, 'authorized_keys');

  exports.init = function(configPath, configData, cb) {
    var cb2, ref, regPath;
    regPath = configData != null ? (ref = configData.server) != null ? ref.registryfile : void 0 : void 0;
    if (regPath == null) {
      return cb(new Error("No registryfile in configuration"));
    }
    cb2 = createAuthkeys(regPath, cb);
    return fs.exists(regPath, function(exists) {
      var data;
      if (exists) {
        return cb2(null, false);
      } else {
        data = {
          maxRegisteredId: 1000,
          maxRegisteredPort: 11080,
          comments: ['Keys used by petit serverd to verify identity and associate', 'Each key is also in the local authorized_keys file'],
          version: 1,
          registeredIds: {}
        };
        return fs.writeFile(regPath, JSON.stringify(data, null, 2), {
          encoding: 'utf8'
        }, function(err) {
          return cb2(err, true);
        });
      }
    });
  };

  createAuthkeys = function(regPath, cb) {
    return function(err, made) {
      if (err != null) {
        return cb(err, made, regPath);
      }
      return fs.exists(akPath, function(exists) {
        if (exists) {
          return cb(null, made, regPath);
        }
        return mkdirp(shPath, 0o700, function(err) {
          if (err != null) {
            return cb(err, made, regPath);
          }
          return fs.writeFile(akPath, '', {
            encoding: 'ascii',
            mode: 0o600
          }, function(err) {
            return cb(err, made, regPath);
          });
        });
      });
    };
  };

  exports.manage = function(configPath, configData, cb) {
    var ref, regPath;
    regPath = configData != null ? (ref = configData.server) != null ? ref.registryfile : void 0 : void 0;
    if (regPath == null) {
      return cb(new Error("No registryfile in configuration"));
    }
    return fs.readFile(regPath, {
      encoding: 'utf8'
    }, function(err, data) {
      if (err != null) {
        return cb(err);
      }
      data = JSON.parse(data);
      return cb(null, createManager(regPath, data));
    });
  };

  createManager = function(regPath, data) {
    var addAlias, addAuthorized, alias, aliases, connected, connecting, connects, disconnected, each, entry, find, findAlias, get, i, isAvailable, isConnected, j, len, len1, ref, ref1, ref2, regId, register, removeAlias, removeAuthorized, sameKey, setVitae, size, status, statusOf, unregister, usage, used, validatePreference, write;
    connects = {};
    aliases = {};
    ref = _.keys(data.registeredIds);
    for (i = 0, len = ref.length; i < len; i++) {
      regId = ref[i];
      entry = data.registeredIds[regId];
      ref2 = (ref1 = entry.aliases) != null ? ref1 : [];
      for (j = 0, len1 = ref2.length; j < len1; j++) {
        alias = ref2[j];
        aliases[alias] = regId;
      }
    }
    write = function(cb) {
      return fs.writeFile(regPath, JSON.stringify(data, null, 2), {
        encoding: 'utf8'
      }, cb);
    };
    status = function(cb) {
      var fpt, pubkey;
      return cb(null, {
        status: 'running',
        connections: (function() {
          var len2, m, ref3, ref4, results;
          ref3 = _.keys(data.registeredIds);
          results = [];
          for (m = 0, len2 = ref3.length; m < len2; m++) {
            regId = ref3[m];
            entry = data.registeredIds[regId];
            pubkey = forge.pki.publicKeyFromPem(entry.key);
            fpt = forge.ssh.getPublicKeyFingerprint(pubkey, {
              encoding: 'hex',
              delimiter: ':'
            });
            results.push({
              registeredId: regId,
              port: entry.port,
              fingerprint: fpt,
              connected: (ref4 = connects[regId]) != null ? ref4 : false,
              aliases: entry.aliases,
              vitae: entry.vitae
            });
          }
          return results;
        })()
      });
    };
    statusOf = function(rid, cb) {
      var fpt, pubkey;
      return cb(null, {
        status: 'running',
        connections: (function() {
          var len2, m, ref3, ref4, results;
          ref3 = [rid];
          results = [];
          for (m = 0, len2 = ref3.length; m < len2; m++) {
            regId = ref3[m];
            entry = data.registeredIds[regId];
            pubkey = forge.pki.publicKeyFromPem(entry.key);
            fpt = forge.ssh.getPublicKeyFingerprint(pubkey, {
              encoding: 'hex',
              delimiter: ':'
            });
            results.push({
              registeredId: regId,
              port: entry.port,
              fingerprint: fpt,
              connected: (ref4 = connects[regId]) != null ? ref4 : false,
              aliases: entry.aliases,
              vitae: entry.vitae
            });
          }
          return results;
        })()
      });
    };
    addAlias = function(regId, alias, cb) {
      if (aliases[alias] != null) {
        return cb(`Alias ${alias} already exists.`);
      }
      if (data.registeredIds[regId] == null) {
        return cb(`No such registeredId ${regId}.`);
      }
      if (data.registeredIds[regId].aliases == null) {
        data.registeredIds[regId].aliases = [];
      }
      data.registeredIds[regId].aliases.push(alias);
      aliases[alias] = regId;
      return write(cb);
    };
    removeAlias = function(regId, alias, cb) {
      if (data.registeredIds[regId] == null) {
        return cb(`No such registeredId ${regId}.`);
      }
      if (data.registeredIds[regId].aliases != null) {
        data.registeredIds[regId].aliases = _.reject(data.registeredIds[regId].aliases, function(a) {
          return a === alias;
        });
      }
      delete aliases[alias];
      return write(cb);
    };
    sameKey = function(regIds, cb) {
      var commonfpt, e, fpt, len2, m, pubkey;
      commonfpt = null;
      for (m = 0, len2 = regIds.length; m < len2; m++) {
        regId = regIds[m];
        if ((entry = data.registeredIds[regId]) == null) {
          return cb(null, false);
        }
        try {
          pubkey = forge.pki.publicKeyFromPem(entry.key);
          fpt = forge.ssh.getPublicKeyFingerprint(pubkey, {
            encoding: 'hex',
            delimiter: ':'
          });
        } catch (error) {
          e = error;
          return cb(e);
        }
        if (commonfpt == null) {
          commonfpt = fpt;
        } else if (commonfpt !== fpt) {
          return cb(null, false);
        }
      }
      return cb(null, true);
    };
    setVitae = function(regId, vitae, cb) {
      if ((entry = data.registeredIds[regId]) == null) {
        return cb(`No such registeredId ${regId}`);
      }
      if (vitae != null) {
        entry.vitae = vitae;
      } else {
        delete entry.vitae;
      }
      return write(cb);
    };
    findAlias = function(alias, cb) {
      return cb(null, aliases[alias]);
    };
    isAvailable = function(regIds) {
      var len2, m, ref3, takens;
      takens = [];
      for (m = 0, len2 = regIds.length; m < len2; m++) {
        regId = regIds[m];
        if (!(((ref3 = connects[regId]) != null ? ref3 : false) === false)) {
          takens.push(regId);
        }
      }
      return takens;
    };
    isConnected = function(regId) {
      var ref3;
      return ((ref3 = connects[regId]) != null ? ref3 : false) === true;
    };
    connected = function(regId) {
      return connects[regId] = true;
    };
    connecting = function(regIds) {
      var len2, m, results;
      results = [];
      for (m = 0, len2 = regIds.length; m < len2; m++) {
        regId = regIds[m];
        results.push(connects[regId] = "connecting");
      }
      return results;
    };
    disconnected = function(regId) {
      return connects[regId] = false;
    };
    size = function(cb) {
      return cb(null, _.keys(data.registeredIds).length);
    };
    get = function(regId, cb) {
      entry = data.registeredIds[regId];
      if ((entry == null) || (entry.port == null) || (entry.key == null)) {
        return cb(new Error(`No such registeredId \"${regId}\".`));
      }
      return cb(null, entry);
    };
    find = function(regIds, cb) {
      var key, len2, m, ports;
      ports = [];
      key = null;
      for (m = 0, len2 = regIds.length; m < len2; m++) {
        regId = regIds[m];
        entry = data.registeredIds[regId];
        if ((entry == null) || (entry.port == null) || (entry.key == null)) {
          return cb(new Error(`No such registeredId \"${regId}\".`));
        }
        if (key == null) {
          key = entry.key;
        }
        ports.push(entry.port);
      }
      return cb(null, key, ports);
    };
    each = function(icb) {
      var fpt, key, len2, m, pubkey, ref3, results;
      ref3 = _.keys(data.registeredIds);
      results = [];
      for (m = 0, len2 = ref3.length; m < len2; m++) {
        key = ref3[m];
        entry = data.registeredIds[key];
        pubkey = forge.pki.publicKeyFromPem(entry.key);
        fpt = forge.ssh.getPublicKeyFingerprint(pubkey, {
          encoding: 'hex',
          delimiter: ':'
        });
        results.push(icb(null, {
          fingerprint: fpt,
          port: entry.port,
          id: key,
          aliases: entry.aliases
        }));
      }
      return results;
    };
    used = function(id, score, data, cb) {
      return cb(null);
    };
    usage = function(id, start, end, cb) {
      return cb(null, [
        {
          msg: "No in-memory usage"
        }
      ]);
    };
    validatePreference = function(preferred) {
      if ((preferred == null) || (preferred.id == null)) {
        return false;
      }
      if (typeof preferred.id !== 'string') {
        return false;
      }
      if (preferred.id.length < 4) {
        return false;
      }
      if (preferred.port != null) {
        if (!_.isInteger(preferred.port)) {
          return false;
        }
        if (preferred.port < 11080) {
          return false;
        }
        if (preferred.port > 91080) {
          return false;
        }
      }
      return true;
    };
    register = function(pubkeypem, preferred, cb) {
      var err, max, pubkey, registeredId, registeredPort, sshkey;
      if ((data.maxRegisteredPort == null) && (data.lastPort != null)) {
        data.maxRegisteredPort = data.lastPort;
      }
      if ((data.maxRegisteredId == null) && (data.lastRegisteredId != null)) {
        data.maxRegisteredId = data.lastRegisteredId;
      }
      if (validatePreference(preferred)) {
        if (!_.some(_.keys(data.registeredIds), function(k) {
          return k === preferred.id;
        })) {
          registeredId = preferred.id;
        }
        if ((preferred.port != null) && !_.some(_.values(data.registeredIds), function(e) {
          return e.port === preferred.port;
        })) {
          registeredPort = preferred.port;
        }
      }
      if (registeredId == null) {
        registeredId = (data.maxRegisteredId + 1).toString();
      }
      if (registeredPort == null) {
        registeredPort = data.maxRegisteredPort + 1;
      }
      try {
        if (pubkeypem != null) {
          pubkey = forge.pki.publicKeyFromPem(pubkeypem);
        }
        if (pubkey != null) {
          sshkey = forge.ssh.publicKeyToOpenSSH(pubkey);
        }
      } catch (error) {
        err = error;
        return cb(err);
      }
      if (sshkey == null) {
        return cb(new Error('Failed public key.'));
      }
      entry = data.registeredIds[registeredId] = {
        port: registeredPort,
        key: pubkeypem
      };
      if (/^\d+$/.test(registeredId)) {
        max = Math.max(data.maxRegisteredId, parseInt(registeredId));
        if (!isNaN(max)) {
          data.maxRegisteredId = max;
        }
      }
      data.maxRegisteredPort = Math.max(data.maxRegisteredPort, registeredPort);
      return addAuthorized(sshkey, registeredPort, function(err) {
        if (err != null) {
          return cb(err);
        }
        return write(function(err) {
          if (err != null) {
            return cb(err);
          }
          return cb(null, registeredId, entry);
        });
      });
    };
    unregister = function(regIds, cb) {
      return async.eachSeries(regIds, function(regId, cb) {
        entry = data.registeredIds[regId];
        if (entry == null) {
          return cb(`No entry found for ${regId}`);
        }
        delete data.registeredIds[regId];
        if (_.keys(data.registeredIds).length === 0) {
          data.maxRegisteredId = 1000;
          data.maxRegisteredPort = 11080;
        }
        return removeAuthorized(entry.port, function(err) {
          if (err != null) {
            return cb(err);
          }
          return write(function(err) {
            if (err != null) {
              return cb(err);
            }
            return cb(null, regId, entry);
          });
        });
      }, cb);
    };
    addAuthorized = function(sshkey, port, cb) {
      return fs.appendFile(akPath, `command=\"echo 'This login is restricted.'\",no-agent-forwarding,no-X11-forwarding,permitopen=\"localhost:${port}\" ${sshkey}\n`, {
        encoding: 'utf8'
      }, cb);
    };
    removeAuthorized = function(port, cb) {
      return fs.readFile(akPath, {
        encoding: 'utf8'
      }, function(err, data) {
        var lines, origLines;
        if (err != null) {
          return cb(err);
        }
        origLines = data.split('\n');
        lines = _.reject(origLines, function(l) {
          return l.indexOf(`localhost:${port}`) > -1;
        });
        return fs.writeFile(akPath, lines.join('\n'), {
          encoding: 'utf8'
        }, cb);
      });
    };
    return {
      setVitae: setVitae,
      findAlias: findAlias,
      addAlias: addAlias,
      removeAlias: removeAlias,
      status: status,
      connected: connected,
      connecting: connecting,
      isAvailable: isAvailable,
      isConnected: isConnected,
      disconnected: disconnected,
      size: size,
      each: each,
      first: first,
      find: find,
      get: get,
      write: write,
      register: register,
      unregister: unregister,
      sameKey: sameKey,
      used: used,
      usage: usage
    };
  };

}).call(this);
