(function() {
  // restapi.coffee    Copyright (c) 2017 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.
  var _, command, config, events, findCounts, findRegisters, findUsage, lock, net, say, scoring, sendTo, util, warn;

  net = require('net');

  _ = require('lodash');

  events = require('events');

  config = require('../config');

  lock = require('../lock');

  util = require('./util');

  say = console.log;

  warn = console.warn;

  exports.start = function(port, opts, output) {
    var allowCrossDomain, app, bodyParser, cmd, express;
    express = require('express');
    bodyParser = require('body-parser');
    allowCrossDomain = function(req, res, next) {
      var ref, ref1;
      res.header('Access-Control-Allow-Credentials', true);
      res.header('Access-Control-Allow-Origin', typeof req.headers === "function" ? req.headers(typeof origin !== "undefined" && origin !== null ? origin : '*') : void 0);
      res.header('Access-Control-Allow-Methods', 'PUT,OPTIONS,GET,DELETE,POST');
      res.header('Access-Control-Allow-Headers', (ref = (ref1 = req.headers) != null ? ref1['Access-Control-Request-Headers'] : void 0) != null ? ref : 'x-requested-with,content-type,authorization,accept');
      res.header('Access-Control-Max-Age', 1728000);
      return next();
    };
    app = express();
    app.use(allowCrossDomain);
    app.use(bodyParser.json({
      limit: '1000kb'
    }));
    app.set('etag', false);
    // put something in the body to let crossDomain happen
    app.options('*', function(req, res, next) {
      return res.send(200);
    });
    // commands
    cmd = command(opts, output);
    app.get('/registry', cmd.registry);
    app.get('/status', cmd.status);
    app.get('/ping', cmd.ping);
    app.get('/usage/:id', cmd.usage);
    app.get('/stats/:id', cmd.stats);
    app.get('/counts/registers', cmd.registers);
    app.get('/counts/registrations', cmd.registrations);
    app.get('/counts/errors', cmd.errors('errors'));
    app.get('/counts/requests', cmd.errors('requests'));
    app.listen(port);
    return output.log('Server REST API started on', port);
  };

  command = function(opts, output) {
    return {
      ping: function(req, res) {
        var configPath, send;
        send = sendTo(res);
        configPath = config.pickConfig(opts.options.config);
        return lock.find('server', configPath, function(err, exists, piddata) {
          if (err != null) {
            return warn(err);
          }
          if (!exists) {
            return send({
              status: 500,
              error: "Server is not running."
            });
          }
          return send(null, {
            success: {
              success: true
            }
          });
        });
      },
      status: function(req, res) {
        var configPath, send;
        send = sendTo(res);
        configPath = config.pickConfig(opts.options.config);
        return lock.find('server', configPath, function(err, exists, piddata) {
          var buf, conn;
          if (err != null) {
            return warn(err);
          }
          if (!exists) {
            return send({
              status: 500,
              error: "Server is not running."
            });
          }
          conn = new net.Socket();
          conn.setEncoding('utf8');
          buf = "";
          conn.on('data', function(data) {
            var c, downs, e, obj, ups;
            try {
              buf += data;
              obj = JSON.parse(buf);
            } catch (error) {
              e = error;
              // wait for more
              say("...retrieving data...");
              return;
            }
            say("conns", obj.connections);
            ups = (function() {
              var i, len, ref, results;
              ref = obj.connections;
              results = [];
              for (i = 0, len = ref.length; i < len; i++) {
                c = ref[i];
                if (c.connected) {
                  results.push(c.registeredId);
                }
              }
              return results;
            })();
            downs = (function() {
              var i, len, ref, results;
              ref = obj.connections;
              results = [];
              for (i = 0, len = ref.length; i < len; i++) {
                c = ref[i];
                if (!c.connected) {
                  results.push(c.registeredId);
                }
              }
              return results;
            })();
            return send(null, {
              success: {
                success: true,
                total: ups.length + downs.length,
                connected: ups,
                disconnected: downs
              }
            });
          });
          return conn.connect(piddata.listen, 'localhost', function() {
            return conn.write('STATUS\x03');
          });
        });
      },
      registry: function(req, res) {
        var configPath, send;
        send = sendTo(res);
        configPath = config.pickConfig(opts.options.config);
        return lock.find('server', configPath, function(err, exists, piddata) {
          var buf, conn;
          if (err != null) {
            return warn(err);
          }
          if (!exists) {
            return send({
              status: 500,
              error: "Server is not running."
            });
          }
          conn = new net.Socket();
          conn.setEncoding('utf8');
          buf = "";
          conn.on('data', function(data) {
            var e, obj;
            try {
              buf += data;
              obj = JSON.parse(buf);
            } catch (error) {
              e = error;
              // wait for more
              say("...retrieving data...");
              return;
            }
            return send(null, {
              success: obj
            });
          });
          return conn.connect(piddata.listen, 'localhost', function() {
            return conn.write('STATUS\x03');
          });
        });
      },
      usage: function(req, res) {
        var end, regid, send, start;
        send = sendTo(res);
        regid = req.params.id;
        start = req.query.start;
        end = req.query.end;
        return config.readConfig(opts.options.config, function(err, configPath, configData) {
          if (err != null) {
            return send(`Error reading config \"${configPath}\": ${err}`);
          }
          return util.verifyConfig(configPath, configData, function(err) {
            var ref;
            if (err != null) {
              return send(`Can't verify server: ${err}`);
            }
            if ((ref = configData.server) != null) {
              ref.configPath = configData.configPath;
            }
            return findUsage(regid, start, end, configData.server, function(err, data) {
              if (err != null) {
                return send(`Error showing usage: ${err}`);
              }
              return send(null, {
                success: data
              });
            });
          });
        });
      },
      stats: function(req, res) {
        var end, regid, send, start;
        send = sendTo(res);
        regid = req.params.id;
        start = req.query.start;
        end = req.query.end;
        return config.readConfig(opts.options.config, function(err, configPath, configData) {
          if (err != null) {
            return send(`Error reading config \"${configPath}\": ${err}`);
          }
          return util.verifyConfig(configPath, configData, function(err) {
            var ref;
            if (err != null) {
              return send(`Can't verify server: ${err}`);
            }
            if ((ref = configData.server) != null) {
              ref.configPath = configData.configPath;
            }
            return findUsage(regid, start, end, configData.server, function(err, data) {
              var errors, successes, unavailables;
              if (err != null) {
                return send(`Error showing usage: ${err}`);
              }
              successes = _.filter(data, {
                success: true
              }).length;
              errors = _.filter(data, {
                reason: 'service error'
              }).length;
              unavailables = _.filter(data, {
                reason: 'service unavailable'
              }).length;
              return send(null, {
                success: {successes, errors, unavailables}
              });
            });
          });
        });
      },
      registrations: function(req, res) {
        var end, send, start;
        send = sendTo(res);
        start = req.query.start;
        end = req.query.end;
        return config.readConfig(opts.options.config, function(err, configPath, configData) {
          if (err != null) {
            return send(`Error reading config \"${configPath}\": ${err}`);
          }
          return util.verifyConfig(configPath, configData, function(err) {
            var ref;
            if (err != null) {
              return send(`Can't verify server: ${err}`);
            }
            if ((ref = configData.server) != null) {
              ref.configPath = configData.configPath;
            }
            return configData.server.registry.size(function(err, sz) {
              if (err != null) {
                return send(`Error finding registrations: ${err}`);
              }
              return send(null, {
                success: {
                  count: sz
                }
              });
            });
          });
        });
      },
      registers: function(req, res) {
        var brand, end, send, start;
        send = sendTo(res);
        start = req.query.start;
        end = req.query.end;
        brand = req.headers['elements-g2c-brand'] || 'ce';
        return config.readConfig(opts.options.config, function(err, configPath, configData) {
          if (err != null) {
            return send(`Error reading config \"${configPath}\": ${err}`);
          }
          return util.verifyConfig(configPath, configData, function(err) {
            var ref;
            if (err != null) {
              return send(`Can't verify server: ${err}`);
            }
            if ((ref = configData.server) != null) {
              ref.configPath = configData.configPath;
            }
            return findRegisters(start, end, brand, configData.server, function(err, count) {
              if (err != null) {
                return send(`Error showing registers: ${err}`);
              }
              return send(null, {
                success: {
                  count: count
                }
              });
            });
          });
        });
      },
      errors: function(type) {
        return function(req, res) {
          var brand, code, end, intype, ref, send, start;
          send = sendTo(res);
          start = req.query.start;
          end = req.query.end;
          brand = req.headers['elements-g2c-brand'] || 'ce';
          intype = (ref = req.query.type) != null ? ref : type;
          code = req.query.code;
          return config.readConfig(opts.options.config, function(err, configPath, configData) {
            if (err != null) {
              return send(`Error reading config \"${configPath}\": ${err}`);
            }
            return util.verifyConfig(configPath, configData, function(err) {
              var ref1;
              if (err != null) {
                return send(`Can't verify server: ${err}`);
              }
              if ((ref1 = configData.server) != null) {
                ref1.configPath = configData.configPath;
              }
              return findCounts(intype, code, start, end, brand, configData.server, function(err, count) {
                if (err != null) {
                  return send(`Error showing usage: ${err}`);
                }
                return send(null, {
                  success: {
                    count: count
                  }
                });
              });
            });
          });
        };
      }
    };
  };

  findUsage = function(regid, start, end, csdata, cb) {
    var ses;
    ses = scoring(start, end);
    return csdata.registry.usage(regid, ses[0], ses[1], cb);
  };

  findRegisters = function(start, end, brand, csdata, cb) {
    var ses;
    ses = scoring(start, end);
    return csdata.registry.registers(brand, ses[0], ses[1], cb);
  };

  findCounts = function(type, code, start, end, brand, csdata, cb) {
    var ses;
    ses = scoring(start, end);
    return csdata.registry.counts(type, code, brand, ses[0], ses[1], cb);
  };

  scoring = function(start, end) {
    var endScore, startScore;
    endScore = end != null ? util.toScore(end) : util.nowScore();
    // default = end - 30 days
    startScore = start != null ? util.toScore(start) : endScore - 2592000000;
    return [startScore, endScore];
  };

  sendTo = function(res) {
    return function(err, item) {
      var ref;
      if (err != null) {
        if (err.status != null) {
          res.status(err.status);
          err = err.error;
        } else {
          res.status(500);
        }
        return res.send({
          message: 'Data Error',
          error: err
        });
      } else {
        if (item == null) {
          return res.status(404).send({
            message: 'Not found'
          });
        } else if (item.failure != null) {
          return res.status((ref = item.status) != null ? ref : 500).send(item.failure);
        } else {
          return res.send(item.success);
        }
      }
    };
  };

}).call(this);
