(function() {
  // server.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // The top-level server daemon execution, which opens a listener port for
  // receiving commands, as well as opens up the backchannel and main
  // http(s) proxy.
  var backtalk, connectBacktalk, connectListener, connectProxy, events, listen, proxy, shutdown, start, waitfor;

  events = require('events');

  listen = require("../listen");

  waitfor = require('../waitfor');

  backtalk = require("./backtalk");

  proxy = require("./proxy");

  exports.create = function() {
    var protocol;
    protocol = new events.EventEmitter();
    protocol.listener = new events.EventEmitter();
    protocol.backchannel = new events.EventEmitter();
    protocol.proxy = new events.EventEmitter();
    protocol._isShutdown = false;
    protocol.end = function() {
      if (protocol._isShutdown) {
        return;
      }
      protocol._isShutdown = true;
      return protocol.emit('shutdown');
    };
    protocol.status = function() {
      return {
        status: 'running'
      };
    };
    protocol.start = function(configData) {
      return start(protocol, configData);
    };
    return protocol;
  };

  start = function(protocol, configData) {
    var waiter;
    protocol.status = function(cb) {
      return configData.registry.status(cb);
    };
    protocol.statusOf = function(regId, cb) {
      return configData.registry.statusOf(regId, cb);
    };
    waiter = waitfor.create(function(isErr) {
      return protocol.emit('close', isErr);
    });
    return connectListener(protocol, configData, waiter);
  };

  connectListener = function(protocol, configData, waiter) {
    var lt;
    lt = listen.create();
    lt.on('error', function(err) {
      protocol.listener.emit('error', err);
      protocol.emit('error', err);
      return waiter.error(err);
    });
    lt.on('close', function(hadErr) {
      protocol.listener.emit('close', hadErr);
      return waiter.close();
    });
    lt.on('connection', function(ltc) {
      return protocol.listener.emit('connection', ltc);
    });
    lt.on('command', function(cmd, ltc) {
      protocol.listener.emit('command', cmd, ltc);
      return protocol.emit('command', cmd, ltc);
    });
    lt.on('listening', function(addr) {
      waiter.open();
      protocol.listener.emit('listening', addr);
      return connectBacktalk(protocol, lt, waiter.branch());
    });
    if (!protocol._isShutdown) {
      protocol.on('shutdown', function() {
        return shutdown(lt);
      });
      return lt.listen(configData, 3016);
    } else {
      return waiter.close();
    }
  };

  connectBacktalk = function(protocol, lt, waiter) {
    var bt;
    bt = backtalk.create();
    bt.on('error', function(err) {
      protocol.backchannel.emit('error', err);
      protocol.emit('error', err);
      return waiter.error(err);
    });
    bt.on('close', function(hadErr) {
      protocol.backchannel.emit('close', hadErr);
      shutdown(lt);
      return waiter.close();
    });
    bt.on('connection', function(btc) {
      return protocol.backchannel.emit('connection', btc);
    });
    bt.on('listening', function(addr) {
      waiter.open();
      protocol.backchannel.emit('listening', addr);
      return connectProxy(protocol, lt, bt, waiter.branch());
    });
    lt.on('close', function() {
      return shutdown(null, bt);
    });
    if (!protocol._isShutdown) {
      protocol.on('shutdown', function() {
        return shutdown(null, bt);
      });
      return bt.listen(lt);
    } else {
      return waiter.close();
    }
  };

  connectProxy = function(protocol, lt, bt, waiter) {
    var px;
    px = proxy.create();
    px.on('error', function(err) {
      protocol.proxy.emit('error', err);
      protocol.emit('error', err);
      return waiter.error(err);
    });
    px.on('close', function(hadErr) {
      protocol.proxy.emit('close', hadErr);
      shutdown(lt, bt);
      return waiter.close();
    });
    px.on('nokey', function() {
      return protocol.proxy.emit('nokey');
    });
    px.on('connection', function(ltc, rid) {
      return protocol.proxy.emit('connection', ltc, rid);
    });
    px.on('listening', function() {
      waiter.open();
      return protocol.proxy.emit('listening');
    });
    lt.on('close', function() {
      return shutdown(null, null, px);
    });
    bt.on('close', function() {
      return shutdown(null, null, px);
    });
    if (!protocol._isShutdown) {
      protocol.on('shutdown', function() {
        return shutdown(null, null, px);
      });
      return px.listen(lt, bt);
    } else {
      return waiter.close();
    }
  };

  shutdown = function(lt, bt, px) {
    if (px != null) {
      px.end();
    }
    if (bt != null) {
      bt.end();
    }
    return lt != null ? lt.end() : void 0;
  };

}).call(this);
